///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"
#include "JetByteTools/CoreTools/ZeroInitialiseExpandableBuffer.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/Win32Tools/SystemUtils.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/FullAddress.h"

#include "JetByteTools/SSPITools/Exception.h"

#include "JetByteTools/SChannelTools/Context.h"
#include "JetByteTools/SChannelTools/Credentials.h"

#include "SocketServer.h"
#include "CommandLine.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/SequencedStreamSocketAllocator.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ToHexString;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CGlobalErrorHandler;
using JetByteTools::Core::ErrorCodeToErrorMessage;

using JetByteTools::Win32::GetComputerName;
using JetByteTools::Win32::GetUserName;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::ZeroByteReadConfiguration;

using JetByteTools::SSPI::SChannel::CCredentials;
using JetByteTools::SSPI::SChannel::CContext;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;
using std::string;

typedef JetByteTools::Core::TZeroInitialiseExpandableBuffer<BYTE> ByteBuffer;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions, SChannel specific...
///////////////////////////////////////////////////////////////////////////////

static CCredentials CreateCredentials(
   const _tstring &storeName,
   const _tstring &certificateName,
   bool useMachineStore,
   bool attemptDefaults,
   DWORD enabledProtocols);

typedef std::vector<PCCERT_CONTEXT> Certificates;

static bool GetCertificates(
   Certificates &certificates,
   const _tstring &storeName,
   const _tstring &serviceName,
   const _tstring &certificateName,
   bool useMachineStore,
   bool useServiceStore);

static void DumpCredentials(
   const CCredentials &credentials);

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   CGlobalErrorHandler errorHandler;

   auto result = 0;

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("SChannelEchoServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CSChannelServerCommandLine commandLine(argc, argv);

         if (commandLine.Parse())
         {
            CCredentials credentials(
               CreateCredentials(
                  commandLine.CertificateStoreName(),
                  commandLine.CertificateName(),
                  commandLine.UseMachineStore(),
                  commandLine.UseDefaultsToLocateCertificate(),
                  commandLine.EnabledProtocols()));

            DumpCredentials(credentials);

            CIOPool ioPool(
               commandLine.NumberOfIOThreads(),
               commandLine.DisplayDebug());

            ioPool.Start();

            CSequencedStreamSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket(),
               commandLine.DisplayDebug());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.GetAllocationFlags(),
               commandLine.DisplayDebug());

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            CSocketServer server(
               credentials,
               commandLine.RequireClientCerts(),
               commandLine.RenegotiateClientCertsSwitch(),
               commandLine.AllowUntrustedRootSwitch(),
               commandLine.IgnoreValidationErrors(),
               address,
               commandLine.ListenBacklog(),
               ioPool,
               socketAllocator,
               bufferAllocator,
               commandLine.RecvBufferSize(),
               commandLine.SendBufferSize(),
               commandLine.GetZeroByteReadConfiguration());

            server.Start();

            server.StartAcceptingConnections();

            shutdownHandler.SetServerControl(server, commandLine.EventNamePrefix());

            shutdownHandler.WaitForShutdownRequest();

            server.WaitForShutdownToComplete();

            ioPool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// SChannel specific stuff.
//
// Note: The following code just sets up example SSL server credentials, it's
// been put together from code samples from MSDN. You should read and understand
// the SChannel documentation and then write some real code to initialise
// your credentials in the way that you require.
///////////////////////////////////////////////////////////////////////////////

static CCredentials CreateCredentials(
   const _tstring &storeName,
   const _tstring &certificateName,
   const bool useMachineStore,
   const bool attemptDefaults,
   DWORD enabledProtocols)
{
   Certificates certificates;

   const _tstring serviceName(_T(""));       // not a service

   const bool useServiceStore = false;

   if (!GetCertificates(certificates, storeName, serviceName, certificateName, useMachineStore, useServiceStore) && attemptDefaults)
   {
      OutputEx(_T("Failed to locate certificate, trying defaults"));

      const _tstring computerName = GetComputerName();

      if (computerName != certificateName || storeName != _T("MY"))
      {
         OutputEx(_T("Try: \"") + computerName + _T("\" in store: \"") + storeName + _T("\""));

         if (!GetCertificates(certificates, storeName, serviceName, computerName, false, false))
         {
            if (!GetCertificates(certificates, storeName, serviceName, computerName, true, false))
            {
               if (!GetCertificates(certificates, storeName, serviceName, computerName, false, true))
               {
                  GetCertificates(certificates, _T("Root"), serviceName, computerName, false, true);
               }
            }
         }
      }

      if (certificates.empty())
      {
         const _tstring userName = GetUserName();

         OutputEx(_T("Try: \"") + userName + _T("\" in store: \"") + storeName + _T("\""));

         if (userName != certificateName || storeName != _T("MY"))
         {
            if (!GetCertificates(certificates, storeName, serviceName, userName, false, false))
            {
               if (!GetCertificates(certificates, storeName, serviceName, userName, true, false))
               {
                  GetCertificates(certificates, storeName, serviceName, computerName, false, true);
               }
            }
         }
      }
   }

   if (certificates.empty())
   {
      throw CException(_T("CreateCredentials()"), _T("No certificates available"));
   }

   CCredentials::Data data;

   data.cCreds = static_cast<DWORD>(certificates.size());
   data.paCred = &(certificates[0]);

   enabledProtocols = 0; // Enable all protocols

   data.grbitEnabledProtocols = enabledProtocols;

   CCredentials credentials(data, SECPKG_CRED_INBOUND);

   for (Certificates::const_iterator it = certificates.begin(), end = certificates.end();
      it != end;
      ++it)
   {
      CertFreeCertificateContext(*it);
   }

   return credentials;
}

#pragma comment(lib, "Crypt32.lib")

static bool CheckCertificate(
   PCCERT_CONTEXT pCertContext)
{
   bool useCertificate = false;

   // Note that for TLS 1.2 you need a certificate that is signed with SHA1 and NOT MD5.
   // If you have an MD5 certificate then TLS 1.2 will not connect and will NOT downgrade
   // to TLS 1.1 or something else...

   DWORD dataSize = 0;

   if (::CertGetCertificateContextProperty(
      pCertContext,
      CERT_SIGN_HASH_CNG_ALG_PROP_ID,
      nullptr,
      &dataSize))
   {
      ByteBuffer buffer(dataSize);

      BYTE *pData = buffer.GetBuffer();

      if (::CertGetCertificateContextProperty(
         pCertContext,
         CERT_SIGN_HASH_CNG_ALG_PROP_ID,
         pData,
         &dataSize))
      {
         const _tstring signHashDetails(CStringConverter::WtoT(reinterpret_cast<wchar_t *>(pData), dataSize / sizeof(wchar_t)));

         OutputEx(_T("Certificate is signed with: ") + signHashDetails);

         //useCertificate = true;

         if (signHashDetails.find(_T("RSA")) == _tstring::npos)
         {
            useCertificate = true;

            OutputEx(_T("Using certificate"));
         }
         else
         {
            OutputEx(_T("NOT Using certificate"));
         }

         if (signHashDetails.find(_T("MD5")) != _tstring::npos)
         {
            OutputEx(_T("MD5 hashes do not work with TLS 1.2 and connection downgrade does NOT occur"));
            OutputEx(_T("This results in failure to connect."));

            throw CException(_T("GetCertificates()"), _T("MD5 certificate will cause problems with TLS 1.2"));
         }
      }
   }

   dataSize = 0;

   if (::CertGetCertificateContextProperty(
      pCertContext,
      CERT_CA_OCSP_AUTHORITY_INFO_ACCESS_PROP_ID,
      0,
      &dataSize))
   {
      ByteBuffer buffer(dataSize);

      BYTE *pData = buffer.GetBuffer();

      if (::CertGetCertificateContextProperty(
         pCertContext,
         CERT_CA_OCSP_AUTHORITY_INFO_ACCESS_PROP_ID,
         pData,
         &dataSize))
      {
         OutputEx(_T("Certificate includes OCSP Authority information"));

         // const _tstring oscpInfo(CStringConverter::WtoT(reinterpret_cast<wchar_t *>(pData), dataSize / sizeof(wchar_t)));
         // ASN.1 string...
      }
   }

   dataSize = 0;

   if (::CertGetCertificateContextProperty(
      pCertContext,
      CERT_CA_DISABLE_CRL_PROP_ID,
      0,
      &dataSize))
   {
      ByteBuffer buffer(dataSize);

      BYTE *pData = buffer.GetBuffer();

      if (::CertGetCertificateContextProperty(
         pCertContext,
         CERT_CA_DISABLE_CRL_PROP_ID,
         pData,
         &dataSize))
      {
         OutputEx(_T("Certificate disables CRL retrieval"));
      }
   }

   return useCertificate;
}

static bool GetCertificates(
   Certificates &certificates,
   const _tstring &storeName,
   const _tstring &serviceName,
   const _tstring &certificateName,
   const bool useMachineStore,
   const bool useServiceStore)
{
   if (storeName.length() == 0)
   {
      throw CException(_T("GetCertificates()"), _T("No certificate store name specified!"));
   }

   if (certificateName.length() == 0)
   {
      throw CException(_T("GetCertificates()"), _T("No certificate name specified!"));
   }

   HCERTSTORE hMyCertStore ;

#ifdef _UNICODE
#define JETBYTE_CERT_STORE_PROV_SYSTEM CERT_STORE_PROV_SYSTEM_W
#else
#define JETBYTE_CERT_STORE_PROV_SYSTEM CERT_STORE_PROV_SYSTEM_A
#endif

   if (useServiceStore)
   {
      const _tstring serviceStoreName = serviceName + _T("\\") + storeName;

      OutputEx(_T("Using service store: ") + serviceStoreName);

      hMyCertStore = CertOpenStore(
         JETBYTE_CERT_STORE_PROV_SYSTEM,
         X509_ASN_ENCODING,
         0,
         CERT_SYSTEM_STORE_SERVICES | CERT_STORE_OPEN_EXISTING_FLAG | CERT_STORE_READONLY_FLAG,
         serviceStoreName.c_str());
   }
   else if (useMachineStore)
   {
      OutputEx(_T("Using machine store"));

      hMyCertStore = CertOpenStore(
         JETBYTE_CERT_STORE_PROV_SYSTEM,
         X509_ASN_ENCODING,
         0,
         CERT_SYSTEM_STORE_LOCAL_MACHINE,
         storeName.c_str());
   }
   else
   {
      OutputEx(_T("Using certificate store: ") + storeName);

      hMyCertStore = CertOpenSystemStore(0, storeName.c_str());
   }

   if (!hMyCertStore)
   {
      const DWORD lastError = GetLastError();

      OutputEx(_T("GetCertificates - CertOpenSystemStore() - failed - ") + ErrorCodeToErrorMessage(lastError));

      return false;
   }

#ifdef _UNICODE
#define JETBYTE_CERT_FIND_SUBJECT_STR CERT_FIND_SUBJECT_STR_W
#else
#define JETBYTE_CERT_FIND_SUBJECT_STR CERT_FIND_SUBJECT_STR_A
#endif

   PCCERT_CONTEXT pCertContext = 0;

   do
   {
      // Find certificate. Note that this just searchs for a certificate
      // that contains the user name somewhere in the subject name.

      pCertContext = CertFindCertificateInStore(
         hMyCertStore,
         X509_ASN_ENCODING,
         0,
         JETBYTE_CERT_FIND_SUBJECT_STR,
         certificateName.c_str(),
         pCertContext);

      if (!pCertContext)
      {
         const DWORD lastError = GetLastError();

         if (CRYPT_E_NOT_FOUND != lastError ||
             certificates.empty())
         {
            OutputEx(_T("GetCertificates - CertFindCertificateInStore() - \"") + certificateName + _T("\" - ") + ErrorCodeToErrorMessage(lastError));
         }
      }
      else
      {
         OutputEx(_T("Found certificate: \"") + certificateName + _T("\""));

         if (CheckCertificate(pCertContext))
         {
            certificates.push_back(CertDuplicateCertificateContext(pCertContext));

            if (useServiceStore && ::IsDebuggerPresent())
            {
               OutputEx(_T("A certificate sourced from the service store will not work if you're not running as a service!"));
               OutputEx(_T("You may get: \"The credentials supplied to the package were not recognized\""));
            }
         }
      }
   }
   while (pCertContext);

   return certificates.size() != 0;
}

static void DumpCredentials(
   const CCredentials &credentials)
{
   ByteBuffer buffer(4096);

   CredHandle hCred = credentials;

   SECURITY_STATUS status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_SUPPORTED_ALGS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query supported algorithms"), status);

   const _SecPkgCred_SupportedAlgs *pAlgs = reinterpret_cast<const _SecPkgCred_SupportedAlgs *>(buffer.GetBuffer());

   for (DWORD i = 0; i < pAlgs->cSupportedAlgs; ++i)
   {
      const _tstring asString = CContext::GetAlgorithmAsString(pAlgs->palgSupportedAlgs[i]);

      if (asString.length())
      {
         OutputEx(ToHexString(pAlgs->palgSupportedAlgs[i]) + _T(" - ") + asString);
      }
      else
      {
         OutputEx(ToHexString(pAlgs->palgSupportedAlgs[i]));
      }
   }

   status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_CIPHER_STRENGTHS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query cipher strengths"), status);

   const _SecPkgCred_CipherStrengths *pCipher = reinterpret_cast<const _SecPkgCred_CipherStrengths *>(buffer.GetBuffer());

   OutputEx(_T("Min cipher strength: ") + ToString(pCipher->dwMinimumCipherStrength));
   OutputEx(_T("Max cipher strength: ") + ToString(pCipher->dwMaximumCipherStrength));

   status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_SUPPORTED_PROTOCOLS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query supported protocols"), status);

   const _SecPkgCred_SupportedProtocols  *pProto = reinterpret_cast<const _SecPkgCred_SupportedProtocols  *>(buffer.GetBuffer());

   OutputEx(JetByteTools::Core::ToHexString(pProto->grbitProtocol) + _T(" - ") + CContext::GetProtocolAsString(pProto->grbitProtocol));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
