///////////////////////////////////////////////////////////////////////////////
// File: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/TChar.h"

#include "JetByteTools/Win32Tools/SystemUtils.h"

#include "JetByteTools/SChannelTools/RemoteCredentialValidation.h"

#pragma hdrstop

#include <iostream>

#include <schannel.h>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToString;

using JetByteTools::Win32::GetComputerName;

using JetByteTools::Socket::ConsecutiveSocketOperations;
using JetByteTools::Socket::PerSocketBufferPoolingLimit;

using JetByteTools::SSPI::SChannel::ValidateRemoteCredentialsAutomatically;
using JetByteTools::SSPI::SChannel::ValidateRemoteCredentialsManually;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Constants
///////////////////////////////////////////////////////////////////////////////

static const _tstring s_serverArgument                   = _T("server");
static const _tstring s_portArgument                     = _T("port");
static const _tstring s_connectionsArgument              = _T("connections");
static const _tstring s_connectionsBatchSizeArgument     = _T("connectionBatchSize");
static const _tstring s_connectionsBatchDelayArgument    = _T("connectionBatchDelay");
static const _tstring s_messagesArgument                 = _T("messages");
static const _tstring s_messageSizeArgument              = _T("messageSize");
static const _tstring s_messageRateArgument              = _T("messageRate");
static const _tstring s_skipSignOnBytesArgument          = _T("skipSignOnBytes");
static const _tstring s_messageFlowTimeoutArgument       = _T("messageFlowTimeout");
static const _tstring s_checkEveryArgument               = _T("checkEvery");
static const _tstring s_readTimeoutArgument              = _T("readTimeout");
static const _tstring s_numberOfIOThreadsArgument        = _T("numberOfIOThreads");

static const _tstring s_storeNameArgument                = _T("storeName");
static const _tstring s_certificateNameArgument          = _T("certificateName");
static const _tstring s_targetNameArgument               = _T("targetName");
static const _tstring s_maxBuffersPooledPerSocket        = _T("perSocketBufferPoolSize");
static const _tstring s_maxConsecutiveOperations         = _T("maxOps");

static const _tstring s_isInteractiveSwitch              = _T("pause");
static const _tstring s_sendAfterRecvSwitch              = _T("sendAfterRecv");
static const _tstring s_syncConnectSwitch                = _T("syncConnect");
static const _tstring s_holdSwitch                       = _T("hold");
static const _tstring s_preallocateSwitch                = _T("preallocate");
static const _tstring s_fragmentsSwitch                  = _T("fragments");
static const _tstring s_multipleMessagesSwitch           = _T("multipleMessages");
static const _tstring s_noConnectionDisplaySwitch        = _T("noConnectionDisplay");

static const _tstring s_useClientCertsSwitch             = _T("useClientCert");
static const _tstring s_useMachineStoreSwitch            = _T("useMachineStore");
static const _tstring s_autoValidateSwitch               = _T("autoValidate");
static const _tstring s_useTargetNameSwitch              = _T("useTargetName");
static const _tstring s_ignoreValidationErrorsSwitch     = _T("ignoreValidationErrors");
static const _tstring s_useSSL2Switch                    = _T("ssl2");
static const _tstring s_useSSL3Switch                    = _T("ssl3");
static const _tstring s_useTLS10Switch                   = _T("tls1");
static const _tstring s_useTLS11Switch                   = _T("tls1.1");
static const _tstring s_useTLS12Switch                   = _T("tls1.2");

///////////////////////////////////////////////////////////////////////////////
// CCommandLine
///////////////////////////////////////////////////////////////////////////////

CCommandLine::CCommandLine()
{
   m_parser.AddArgument(s_serverArgument);
   m_parser.AddArgument(s_portArgument);
   m_parser.AddArgument(s_connectionsArgument);
   m_parser.AddArgument(s_connectionsBatchSizeArgument);
   m_parser.AddArgument(s_connectionsBatchDelayArgument);
   m_parser.AddArgument(s_messagesArgument);
   m_parser.AddArgument(s_messageSizeArgument);
   m_parser.AddArgument(s_messageRateArgument);
   m_parser.AddArgument(s_skipSignOnBytesArgument);
   m_parser.AddArgument(s_messageFlowTimeoutArgument);
   m_parser.AddArgument(s_checkEveryArgument);
   m_parser.AddArgument(s_readTimeoutArgument);
   m_parser.AddArgument(s_numberOfIOThreadsArgument);

   m_parser.AddArgument(s_storeNameArgument);
   m_parser.AddArgument(s_certificateNameArgument);
   m_parser.AddArgument(s_targetNameArgument);
   m_parser.AddArgument(s_maxBuffersPooledPerSocket);
   m_parser.AddArgument(s_maxConsecutiveOperations);

   m_parser.AddSwitch(s_isInteractiveSwitch);
   m_parser.AddSwitch(s_sendAfterRecvSwitch);
   m_parser.AddSwitch(s_syncConnectSwitch);
   m_parser.AddSwitch(s_holdSwitch);
   m_parser.AddSwitch(s_preallocateSwitch);
   m_parser.AddSwitch(s_fragmentsSwitch);
   m_parser.AddSwitch(s_multipleMessagesSwitch);
   m_parser.AddSwitch(s_noConnectionDisplaySwitch);

   m_parser.AddSwitch(s_useClientCertsSwitch);
   m_parser.AddSwitch(s_useMachineStoreSwitch);
   m_parser.AddSwitch(s_autoValidateSwitch);
   m_parser.AddSwitch(s_useTargetNameSwitch);
   m_parser.AddSwitch(s_ignoreValidationErrorsSwitch);
   m_parser.AddSwitch(s_useSSL2Switch);
   m_parser.AddSwitch(s_useSSL3Switch);
   m_parser.AddSwitch(s_useTLS10Switch);
   m_parser.AddSwitch(s_useTLS11Switch);
   m_parser.AddSwitch(s_useTLS12Switch);
}

bool CCommandLine::Parse(
   const DWORD argc,
   const char **argv)
{
   _tstring message;

   bool ok = m_parser.Parse(argc, argv, message);

   if (ok)
   {
      ok = m_parser.IsSet(s_serverArgument) && m_parser.IsSet(s_portArgument);
   }

   if (ok)
   {
      if (m_parser.IsSet(s_messageSizeArgument) || m_parser.IsSet(s_messageRateArgument))
      {
         if (0 == Messages())
         {
            message = _T("Error: if -messageSize or -messageRate is set then -messages must be set to non zero");

            ok = false;
         }
      }
   }

   if (ok)
   {
      if (!m_parser.IsSet(s_holdSwitch) && m_parser.IsSet(s_isInteractiveSwitch))
      {
         message = _T("Error: -pause is only appropriate with -hold");

         ok = false;
      }
   }

   if (ok)
   {
      if (!m_parser.IsSet(s_messageFlowTimeoutArgument) && m_parser.IsSet(s_checkEveryArgument))
      {
         message = _T("Error: -checkEvery is only appropriate with -messageFlowTimeout");

         ok = false;
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxBuffersPooledPerSocket))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxBuffersPooledPerSocket).c_str());

         if (value > std::numeric_limits<PerSocketBufferPoolingLimit>::max())
         {
            message = _T("Error: -perSocketBufferPoolSize value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (ok)
   {
      if (m_parser.IsSet(s_maxConsecutiveOperations))
      {
         const unsigned int value = _ttoi(m_parser.GetValueAsString(s_maxConsecutiveOperations).c_str());
         
         if (value > std::numeric_limits<ConsecutiveSocketOperations>::max())
         {
            message = _T("Error: -maxOps value \"") + ToString(value) + _T("\" out of range, maximum is 255");

            ok = false;
         }
      }
   }

   if (!ok)
   {
      ShowHelp(message);
   }

   return ok ;
}

_tstring CCommandLine::Server() const
{
   return m_parser.GetValueAsString(s_serverArgument);
}

unsigned short CCommandLine::Port() const
{
   return static_cast<unsigned short>(_ttol(m_parser.GetValueAsString(s_portArgument).c_str()));
}

CCommandLine::Count CCommandLine::SkipSignOnMessageBytes() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_skipSignOnBytesArgument, _T("0")).c_str());
}

CCommandLine::Count CCommandLine::Connections() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_connectionsArgument, _T("1000")).c_str());
}

CCommandLine::Count CCommandLine::ConnectionBatchSize() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_connectionsBatchSizeArgument, _T("0")).c_str());
}

Milliseconds CCommandLine::ConnectionBatchPause() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_connectionsBatchDelayArgument, _T("0")).c_str());
}

CCommandLine::Count CCommandLine::Messages() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_messagesArgument, _T("0")).c_str());
}

CCommandLine::Count CCommandLine::MessageSize() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_messageSizeArgument, _T("1024")).c_str());
}

Milliseconds CCommandLine::MessageRate() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_messageRateArgument, _T("1000")).c_str());
}

bool CCommandLine::IsInteractive() const
{
   return m_parser.IsSet(s_isInteractiveSwitch);
}

bool CCommandLine::SendAfterRecv() const
{
   return m_parser.IsSet(s_sendAfterRecvSwitch);
}

bool CCommandLine::SyncConnect() const
{
   return m_parser.IsSet(s_syncConnectSwitch);
}

bool CCommandLine::HoldConnections() const
{
   return m_parser.IsSet(s_holdSwitch);
}

bool CCommandLine::PreAllocate() const
{
   return m_parser.IsSet(s_preallocateSwitch);
}

CCommandLine::Count CCommandLine::NumberOfIOThreads() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_numberOfIOThreadsArgument, _T("0")).c_str());
}

Milliseconds CCommandLine::ReadTimeout() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_readTimeoutArgument, _T("0")).c_str());
}

bool CCommandLine::Fragments() const
{
   return m_parser.IsSet(s_fragmentsSwitch);
}

bool CCommandLine::MultipleMessages() const
{
   return m_parser.IsSet(s_multipleMessagesSwitch);
}

Milliseconds CCommandLine::MessageFlowTimeout() const
{
   if (m_parser.IsSet(s_messageFlowTimeoutArgument))
   {
      return _ttol(m_parser.GetValueAsString(s_messageFlowTimeoutArgument).c_str());
   }

   return INFINITE;
}

Milliseconds CCommandLine::CheckEvery() const
{
   if (m_parser.IsSet(s_checkEveryArgument))
   {
      return _ttol(m_parser.GetValueAsString(s_checkEveryArgument).c_str());
   }

   return INFINITE;
}

bool CCommandLine::DisplayConnectionBatching() const
{
   return !m_parser.IsSet(s_noConnectionDisplaySwitch);
}

bool CCommandLine::UseClientCerts() const
{
   return m_parser.IsSet(s_useClientCertsSwitch);
}

bool CCommandLine::UseMachineStore() const
{
   return m_parser.IsSet(s_useMachineStoreSwitch);
}

_tstring CCommandLine::CertificateStoreName() const
{
   return m_parser.GetOptionalValueAsString(s_storeNameArgument, _T("MY"));
}

_tstring CCommandLine::CertificateName() const
{
   return m_parser.GetOptionalValueAsString(s_certificateNameArgument, GetComputerName());
}

bool CCommandLine::UseTargetName() const
{
   return m_parser.IsSet(s_useTargetNameSwitch);
}

_tstring CCommandLine::TargetName() const
{
   return m_parser.GetOptionalValueAsString(s_targetNameArgument, GetComputerName());
}

JetByteTools::SSPI::SChannel::RemoteCredentialValidation CCommandLine::RemoteCredentialValidation() const
{
   if (m_parser.IsSet(s_autoValidateSwitch))
   {
      return ValidateRemoteCredentialsAutomatically;
   }

   return ValidateRemoteCredentialsManually;
}

bool CCommandLine::IgnoreValidationErrors() const
{
   return m_parser.IsSet(s_ignoreValidationErrorsSwitch);
}

PerSocketBufferPoolingLimit CCommandLine::MaxBuffersPooledPerSocket() const
{
   return static_cast<PerSocketBufferPoolingLimit>(_ttoi(m_parser.GetOptionalValueAsString(s_maxBuffersPooledPerSocket, _T("10")).c_str()));
}

ConsecutiveSocketOperations CCommandLine::MaxConsecutiveOperationsOnOneSocket() const
{
   return static_cast<ConsecutiveSocketOperations>(_ttoi(m_parser.GetOptionalValueAsString(s_maxConsecutiveOperations, _T("10")).c_str()));
}

DWORD CCommandLine::EnabledProtocols() const
{
   DWORD enabledProtocols = 0;      // default to all!

   if (m_parser.IsSet(s_useSSL2Switch))
   {
      enabledProtocols |= SP_PROT_SSL2;
   }

   if (m_parser.IsSet(s_useSSL3Switch))
   {
      enabledProtocols |= SP_PROT_SSL3;
   }

   if (m_parser.IsSet(s_useTLS10Switch))
   {
      enabledProtocols |= SP_PROT_TLS1_0;
   }

   if (m_parser.IsSet(s_useTLS11Switch))
   {
      enabledProtocols |= SP_PROT_TLS1_1;
   }

   if (m_parser.IsSet(s_useTLS12Switch))
   {
      enabledProtocols |= SP_PROT_TLS1_2;
   }

   return enabledProtocols;
}

void CCommandLine::ShowHelp(
   const _tstring &message)
{
   cout << endl;
   cout << "SChannelEchoServerTest: v7.0 - Copyright (c) 2015 JetByte Limited" << endl;

   if (message.length() != 0)
   {
      cout << endl;
      cout << "Error: " << CStringConverter::TtoA(message) << endl;
      cout << endl;
      cout << "In command line: \"" << CStringConverter::TtoA(m_parser.GetParsedCommandLine()) << "\"" << endl;
      cout << endl;
   }

   cout << "Usage: SChannelEchoServerTest -server xxx.xxx.xxx.xxx -port xxxx" << endl;
   cout << endl;
   cout << "Command line parameters:" << endl;
   cout << " r -server                  The server address to connect to (dotted ip)." << endl;
   cout << " r -port                    The port to connect to." << endl;
   cout << " o -connections             The number of connections to create." << endl;
   cout << "                            Defaults to 1000" << endl;
   cout << " o -connectionBatchSize     Batch connections in groups of X size." << endl;
   cout << "                            Defaults to 0 (no batching)" << endl;
   cout << " o -connectionBatchDelay    Delay for Y milliseconds between each batch." << endl;
   cout << "                            Defaults to 0 (no delay)" << endl;
   cout << " o -messages                Number of messages to send on each connection." << endl;
   cout << "                            Defaults to 0 (don't send any data)" << endl;
   cout << " o -messageSize             Size of each message." << endl;
   cout << "                            Defaults to 1024" << endl;
   cout << " o -messageRate             Delay for Y milliseconds between each message." << endl;
   cout << "                            Defaults to 1000" << endl;
   cout << " o -sendAfterRecv           Wait for echo of previous message before starting" << endl;
   cout << "                            timer for next send (reduce pace to server speed)." << endl;
   cout << " o -syncConnect             Use synchronous connect rather than ConnectEx()." << endl;
   cout << " o -hold                    Hold all connections open until test completes." << endl;
   cout << " o -pause                   Wait for a key press before closing connections." << endl;
   cout << " o -preallocate             Preallocate sockets and buffers before connecting." << endl;
   cout << " o -skipSignOnBytes         Ignores the first x bytes that the server sends." << endl;
   cout << "                            This can be used to ignore a welcome message." << endl;
   cout << " o -messageFlowTimeout      Time in milliseconds that is allowed for the test" << endl;
   cout << "                            to complete once all connections are established." << endl;
   cout << "                            Defaults to no timeout, the test doesn't complete" << endl;
   cout << "                            until the data flow completes." << endl;
   cout << " o -noConnectionDisplay     Do not display connection batching information." << endl;
   cout << " o -checkEvery              Time in milliseconds between test status displays." << endl;
   cout << "                            Defaults to never displaying a status display." << endl;
   cout << "                            Status displays are only displayed if" << endl;
   cout << "                            -messageFlowTimeout has been set." << endl;
   cout << " o -readTimeout             Messages must be echoed back within X milliseconds or" << endl;
   cout << "                            the connection is terminated. Defaults to no timeout." << endl;
   cout << " o -fragments               " << endl;
   cout << " o -multipleMessages        " << endl;
   cout << " o -useClientCert" << endl;
   cout << " o -useMachineStore" << endl;
   cout << " o -storeName               Defaults to MY." << endl;
   cout << " o -certificateName         Defaults to the name of the computer" << endl;
   cout << " o -useTargetName           When unset, uses the target address" << endl;
   cout << "                            as the target name for certificate validation." << endl;
   cout << " o -targetName              Defaults to the name of the computer. Only used" << endl;
   cout << "                            if -useTargetName is specified." << endl;
   cout << " o -autoValidate            SChannel automatically validates remote certificates." << endl;
   cout << " o -ignoreValidationErrors  Only valid if -autoValidate is NOT set. Ignores certificate" << endl;
   cout << "                            validation errors. Use with care." << endl;
   cout << " o -perSocketBufferPoolSize The number of I/O buffers pooled in the socket." << endl;
   cout << "                            Defaults to 10." << endl;
   cout << " o -maxOps                  The maximum number of consecutive operations" << endl;
   cout << "                            per socket. Defaults to 10." << endl;
   cout << endl;
   cout << " Enabled protocols" << endl;
   cout << " By default all supported protocols are enabled. If you specify any of the" << endl;
   cout << " command line switches below then ONLY the protocols specified will be" << endl;
   cout << " enabled. So if you want JUST TLS1.1 and TLS1.2 you need to specify" << endl;
   cout << " -tls1.1 -tls1.2 on the command line." << endl;
   cout << endl;
   cout << " o -ssl2                    Enable SSL 2" << endl;
   cout << " o -ssl3                    Enable SSL 3" << endl;
   cout << " o -tls1                    Enable TLS 1.0" << endl;
   cout << " o -tls1.1                  Enable TLS 1.1" << endl;
   cout << " o -tls1.2                  Enable TLS 1.2" << endl;
   cout << endl;
   cout << " r = required, o = optional" << endl;
   cout << endl;
   cout << "See http://www.serverframework.com for more details." << endl;
}

CCommandLine::Count CCommandLine::StringToCount(
   const _tstring &stringValue)
{
   const size_t value = _ttol(stringValue.c_str());

   if (value > static_cast<size_t>(std::numeric_limits<CCommandLine::Count>::max()))
   {
      throw CException(_T("CCommandLine::StringToCount()"), _T("Value too big"));
   }

   return static_cast<Count>(value);
}

///////////////////////////////////////////////////////////////////////////////
// End of file: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
