///////////////////////////////////////////////////////////////////////////////
// File: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "StreamSocketConnectionManager.h"
#include "TestMessages.h"

#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/WaitableCounter.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ToHexString;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ErrorCodeToErrorMessageA;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::ConnectionClosureReason;
using JetByteTools::Socket::ConnectionAbortedMask;
using JetByteTools::Socket::GetClosureReasonAsString;

using JetByteTools::SSPI::SChannel::CCredentials;
using JetByteTools::SSPI::SChannel::CContext;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// CStreamSocketConnectionManager
///////////////////////////////////////////////////////////////////////////////

CStreamSocketConnectionManager::CStreamSocketConnectionManager(
   CCredentials &sslContext,
   const Milliseconds readTimeout,
   const bool ignoreValidationErrors,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator)
   :  JetByteTools::SSPI::SChannel::CStreamSocketConnectionManager(sslContext, *this, pool, socketAllocator, bufferAllocator),
      m_timerHandleIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_timerHandleIndex"))),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_userDataIndex"))),
      m_timeoutFilter(*this, m_timerQueue),
      m_readTimeout(readTimeout),
      m_ignoreValidationErrors(ignoreValidationErrors)
{

}

CStreamSocketConnectionManager::~CStreamSocketConnectionManager()
{
   WaitForShutdownToComplete();
}

void CStreamSocketConnectionManager::OnPreOutgoingConnect(
   IIndexedOpaqueUserData &userData,
   const IAddress & /*address*/,
   const void *pUserData)
{
   const auto pTestMessages = const_cast<CTestMessages *>(static_cast<const CTestMessages*>(pUserData));

   userData.SetUserPointer(m_userDataIndex, pTestMessages);
}

void CStreamSocketConnectionManager::OnSecureConnectionEstablished(
   IStreamSocket & /*socket*/,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnSecureConnectionEstablished")));
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnConnectionEstablished")));

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   pTestMessages->ConnectionComplete(true);

   const auto timerHandle = m_timerQueue.CreateTimer();

   socket.SetUserData(m_timerHandleIndex, timerHandle);

   if (m_readTimeout != 0)
   {
      m_timeoutFilter.SetReadTimeout(socket, m_readTimeout, *this, 1);
   }

   socket.TryRead();

   SendData(socket);
}

void CStreamSocketConnectionManager::OnOutgoingConnectionFailed(
   IStreamSocket &socket,
   const IAddress & /*address*/,
   const DWORD lastError)
{
   {
      ILockableObject::Owner lock(m_lock);

      m_connectionErrors[lastError]++;
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   pTestMessages->ConnectionComplete(false);

   socket.SetUserPointer(m_userDataIndex, nullptr);
}

void CStreamSocketConnectionManager::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   DEBUG_ONLY(Output(_T("OnSocketReleased")));

   const auto timerHandle = userData.GetUserData(m_timerHandleIndex);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      if (m_timerQueue.DestroyTimer(timerHandle))
      {
         OnError(_T("CStreamSocketConnectionManager::OnSocketReleased() - timer was set when destroyed, socket ref count bug!"));
      }
   }
}

void CStreamSocketConnectionManager::OnSecureSocketReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   DEBUG_ONLY(Output(_T("OnSecureSocketReleased")));
}

void CStreamSocketConnectionManager::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   ProcessData(socket, buffer);

   socket.TryRead(buffer);
}

void CStreamSocketConnectionManager::OnConnectionReset(
   IStreamSocket & /*socket*/,
   const DWORD lastError)
{
   ILockableObject::Owner lock(m_lock);

   m_connectionResetErrors[lastError]++;
}

void CStreamSocketConnectionManager::OnError(
   const _tstring &message)
{
   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnCertificateError(
   const _tstring &message)
{
   ILockableObject::Owner lock(m_lock);

   m_certificateErrors[message]++;
}

void CStreamSocketConnectionManager::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason reason)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      m_timerQueue.DestroyTimerWithRefCountedUserData(timerHandle, &socket);
   }

   if (reason & ConnectionAbortedMask)
   {
      OnError(GetClosureReasonAsString(reason));
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      pTestMessages->MessagesComplete(!(reason & ConnectionAbortedMask));
   }
}

void CStreamSocketConnectionManager::ProcessData(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      try
      {
         auto messageComplete = false;

         if (pTestMessages->ValidateReceivedData(buffer->GetMemory(), buffer->GetUsed(), messageComplete))
         {
            if (messageComplete && pTestMessages->SendAfterRecv())
            {
               SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
            }
         }
         else if (messageComplete)
         {
            const auto holdConnections = pTestMessages->HoldConnections();

            if (!holdConnections)
            {
               socket.Shutdown(ShutdownSend);
            }
            else
            {
               pTestMessages->MessagesComplete(true);
            }
         }
      }
      catch (CException &e)
      {
         ILockableObject::Owner lock(m_lock);

         m_errors[e.GetWhat()]++;

         pTestMessages->MessagesComplete(false);

         throw;
      }
   }

   buffer->Empty();
}

void CStreamSocketConnectionManager::OnTimer(
   IQueueTimers::Timer::UserData userData)
{
   auto &socket = *reinterpret_cast<IStreamSocket *>(userData);

   SendData(socket);

   socket.Release();
}

void CStreamSocketConnectionManager::OnTimer(
   IStreamSocket &socket,
   TimerCallback::UserData /*userData*/)
{
   OnError(_T("Read timeout"));

   socket.AbortConnection();
}

void CStreamSocketConnectionManager::SendData(
   IStreamSocket &socket)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      if (pTestMessages->WriteSome(socket))
      {
         if (!pTestMessages->SendAfterRecv())
         {
            SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
         }
      }
      else if (!pTestMessages->HoldConnections())
      {
         socket.Shutdown(ShutdownSend);
      }
   }
}

void CStreamSocketConnectionManager::SetSendTimer(
   IStreamSocket &socket,
   const Milliseconds timeout)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   m_timerQueue.SetTimerWithRefCountedUserData(timerHandle, *this, timeout, &socket);
}

bool CStreamSocketConnectionManager::DumpErrors() const
{
   auto ok = true;

   if (!DumpErrors(_T("Connection establishment errors"), m_connectionErrors))
   {
      ok = false;
   }

   if (!DumpErrors(_T("Connections reset"), m_connectionResetErrors))
   {
      ok = false;
   }

   if (!m_errors.empty())
   {
      ok = false;

      cout << "Other errors" << endl;

      for (const auto &error : m_errors)
      {
         cout << ToStringA(error.second) << " - " << CStringConverter::TtoA(error.first) << endl;
      }
   }

   if (!m_certificateErrors.empty())
   {
      if (!m_ignoreValidationErrors)
      {
         ok = false;
      }

      cout << "Certificate errors" << endl;

      for (const auto &certificateError : m_certificateErrors)
      {
         cout << ToStringA(certificateError.second) << " - " << CStringConverter::TtoA(certificateError.first) << endl;
      }
   }

   return ok;
}

bool CStreamSocketConnectionManager::DumpErrors(
   const _tstring &errorType,
   const LastErrors &lastErrors) const
{
   auto ok = true;

   if (!lastErrors.empty())
   {
      ok = false;

      cout << CStringConverter::TtoA(errorType) << endl;

      for (const auto &lastError : lastErrors)
      {
         cout << ToStringA(lastError.second) << " - " << ErrorCodeToErrorMessageA(lastError.first) << endl;
      }
   }

   return ok;
}

void CStreamSocketConnectionManager::FatalError(
   IStreamSocket &socket,
   const SECURITY_STATUS status)
{
   OnConnectionReset(socket, status);
}

void CStreamSocketConnectionManager::OnContextComplete(
   IStreamSocket & /*socket*/,
   const CContext &context)
{
   Output(_T("CSocketServer::OnContextComplete()"));

   SecPkgContext_ConnectionInfo info;

   context.QueryContextAttributes(info);

   Output(_T("Connection information"));
   Output(_T("       Attributes : ") + CContext::GetAttributesAsString(context.GetAttributes())  + _T(" [") + ToHexString(context.GetAttributes()) + _T("]"));
   Output(_T("         Protocol : ") + CContext::GetProtocolAsString(info.dwProtocol) + _T(" [") + ToHexString(info.dwProtocol) + _T("]"));
   Output(_T("           Cipher : ") + CContext::GetAlgorithmAsString(info.aiCipher) + _T(" [") + ToHexString(info.aiCipher) + _T("]"));
   Output(_T("  Cipher Strength : ") + ToString(info.dwCipherStrength));
   Output(_T("             Hash : ") + CContext::GetAlgorithmAsString(info.aiHash) + _T(" [") + ToHexString(info.aiHash) + _T("]"));
   Output(_T("    Hash Strength : ") + ToString(info.dwHashStrength));
   Output(_T("         Exchange : ") + CContext::GetAlgorithmAsString(info.aiExch) + _T(" [") + ToHexString(info.aiExch) + _T("]"));
   Output(_T("Exchange Strength : ") + ToString(info.dwExchStrength));
}

bool CStreamSocketConnectionManager::Verify(
   IStreamSocket & /*socket*/,
   const _tstring &targetName,
   const CContext &context)
{
   bool ok = true;

   Output(_T("CStreamSocketConnectionManager::Verify() - do manual certificate verification here"));

   CERT_CONTEXT *pCertContext = context.GetRemoteCertificate();

   if (pCertContext)
   {
      DisplayCertChain(pCertContext);

      // Attempt to validate server certificate.

      const SECURITY_STATUS status = VerifyServerCertificate(
         pCertContext,
         targetName);

      if (status != SEC_E_OK)
      {
         // The server certificate did not validate correctly. At this
         // point, we cannot tell if we are connecting to the correct
         // server, or if we are connecting to a "man in the middle"
         // attack server.

         // It is therefore best if we abort the connection.

         if (m_ignoreValidationErrors)
         {
            OnCertificateError(_T("Warning ") + ErrorCodeToErrorMessage(status));
         }
         else
         {
            OnCertificateError(_T("Error ") + ErrorCodeToErrorMessage(status));

            ok = false;
         }
      }

      ::CertFreeCertificateContext(pCertContext);
   }
   else
   {
      OnCertificateError(_T("Server has no certificate"));

      ok = false;
   }

   return ok;
}

void CStreamSocketConnectionManager::DisplayCertChain(
   const CERT_CONTEXT *pServerCert)
{
   if (pServerCert)
   {
      TCHAR szName[1000];

      // display leaf name

      if (!::CertNameToStr(
         pServerCert->dwCertEncodingType,
         &pServerCert->pCertInfo->Subject,
         CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
         szName, sizeof(szName)))
      {
         const DWORD lastError = GetLastError();

         OnCertificateError(_T("DisplayCertChain - Error building subject name: ") + ErrorCodeToErrorMessage(lastError));
      }

      Output(_T("Server subject: ") + _tstring(szName));

      if (!::CertNameToStr(
         pServerCert->dwCertEncodingType,
         &pServerCert->pCertInfo->Issuer,
         CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
         szName, sizeof(szName)))
      {
         const DWORD lastError = GetLastError();

         OnCertificateError(_T("DisplayCertChain - Error building issuer name: ") + ErrorCodeToErrorMessage(lastError));
      }

      Output(_T("Server issuer: ") + _tstring(szName));

      // display certificate chain

      const CERT_CONTEXT *pCurrentCert = pServerCert;

      while (pCurrentCert)
      {
         DWORD dwVerificationFlags = 0;

         const CERT_CONTEXT *pIssuerCert = ::CertGetIssuerCertificateFromStore(
            pServerCert->hCertStore,
            pCurrentCert,
            nullptr,
            &dwVerificationFlags);

         if (!pIssuerCert)
         {
            if (pCurrentCert != pServerCert)
            {
               CertFreeCertificateContext(pCurrentCert);
            }
            break;
         }

         if (!::CertNameToStr(
            pIssuerCert->dwCertEncodingType,
            &pIssuerCert->pCertInfo->Subject,
            CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
            szName, sizeof(szName)))
         {
            const DWORD lastError = GetLastError();

            OnCertificateError(_T("DisplayCertChain - Error building subject name: ") + ErrorCodeToErrorMessage(lastError));
         }

         Output(_T("CA subject: ") + _tstring(szName));

         if (!CertNameToStr(pIssuerCert->dwCertEncodingType,
            &pIssuerCert->pCertInfo->Issuer,
            CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
            szName, sizeof(szName)))
        {
            const DWORD lastError = GetLastError();

            OnCertificateError(_T("DisplayCertChain - Error building issuer name: ") + ErrorCodeToErrorMessage(lastError));
         }

         Output(_T("CA issuer: ") + _tstring(szName));

         if (pCurrentCert != pServerCert)
         {
            CertFreeCertificateContext(pCurrentCert);
         }

         pCurrentCert = pIssuerCert;
         pIssuerCert = nullptr;
      }
   }
}

HRESULT CStreamSocketConnectionManager::VerifyServerCertificate(
   const CERT_CONTEXT *pServerCert,
   const _tstring &serverName)
{
   if (!pServerCert || serverName.length() == 0)
   {
      return SEC_E_WRONG_PRINCIPAL;
   }

   //
   // Build certificate chain.
   //

   CERT_CHAIN_PARA  chainPara;

   LPSTR rgszUsages[] =
   {
      const_cast<LPSTR>(szOID_PKIX_KP_SERVER_AUTH),
      const_cast<LPSTR>(szOID_SERVER_GATED_CRYPTO),
      const_cast<LPSTR>(szOID_SGC_NETSCAPE)
   };

   const DWORD cUsages = sizeof(rgszUsages) / sizeof(LPSTR);

   ZeroMemory(&chainPara, sizeof(chainPara));
   chainPara.cbSize = sizeof(chainPara);
   chainPara.RequestedUsage.dwType = USAGE_MATCH_TYPE_OR;
   chainPara.RequestedUsage.Usage.cUsageIdentifier = cUsages;
   chainPara.RequestedUsage.Usage.rgpszUsageIdentifier = rgszUsages;

   const CERT_CHAIN_CONTEXT *pChainContext = nullptr;

   if (!::CertGetCertificateChain(
      nullptr,
      pServerCert,
      nullptr,
      pServerCert->hCertStore,
      &chainPara,
      0,
      nullptr,
      &pChainContext))
   {
      return GetLastError();
   }

   //
   // Validate certificate chain.
   //

   const std::wstring wideServerName = CStringConverter::TtoW(serverName);

   HTTPSPolicyCallbackData  polHttps;

   const DWORD dwCertFlags = 0;

   ZeroMemory(&polHttps, sizeof(HTTPSPolicyCallbackData));
   polHttps.cbStruct = sizeof(HTTPSPolicyCallbackData);
   polHttps.dwAuthType = AUTHTYPE_SERVER;
   polHttps.fdwChecks = dwCertFlags;
   polHttps.pwszServerName = const_cast<wchar_t *>(wideServerName.c_str());

   CERT_CHAIN_POLICY_PARA   policyPara;

   memset(&policyPara, 0, sizeof(policyPara));
   policyPara.cbSize = sizeof(policyPara);
   policyPara.pvExtraPolicyPara = &polHttps;

   CERT_CHAIN_POLICY_STATUS policyStatus;

   memset(&policyStatus, 0, sizeof(policyStatus));
   policyStatus.cbSize = sizeof(policyStatus);

   if (!::CertVerifyCertificateChainPolicy(
      CERT_CHAIN_POLICY_SSL,
      pChainContext,
      &policyPara,
      &policyStatus))
   {
      return GetLastError();
   }

   DWORD status = SEC_E_OK;

   if (policyStatus.dwError)
   {
      status = policyStatus.dwError;
   }

   if (pChainContext)
   {
      ::CertFreeCertificateChain(pChainContext);
   }

   return status;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
