///////////////////////////////////////////////////////////////////////////////
// File: TestMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "StreamSocketConnectionManager.h"
#include "TestMessages.h"
#include "TestMessage.h"
#include "MessageCollection.h"
#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/PerformanceCounter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"
#include "JetByteTools/CoreTools/ZeroInitialiseExpandableBuffer.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/ManualResetEvent.h"
#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"

#include "JetByteTools/IOTools/IOPool.h"
#include "JetByteTools/IOTools/BufferAllocator.h"
#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"

#include "JetByteTools/SSPITools/Exception.h"

#include "JetByteTools/SChannelTools/Credentials.h"

#pragma hdrstop

#include <algorithm>    // for min

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::CException;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::ToHexString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CPerformanceCounter;
using JetByteTools::Core::CGlobalErrorHandler;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::ErrorCodeToErrorMessage;

using JetByteTools::Win32::CManualResetEvent;
using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;

using JetByteTools::IO::CPendingWriteLimiter;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CUsesWinsock;
using JetByteTools::Socket::CSmartStreamSocket;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;

using JetByteTools::IO::CIOPool;
using JetByteTools::IO::CBufferAllocator;
using JetByteTools::IO::CAsyncFileLog;

using JetByteTools::SSPI::SChannel::CCredentials;
using JetByteTools::SSPI::SChannel::RemoteCredentialValidation;
using JetByteTools::SSPI::SChannel::CContext;

typedef JetByteTools::Core::TZeroInitialiseExpandableBuffer<BYTE> ByteBuffer;

using std::cout;
using std::cin;
using std::endl;
using std::min;
using std::string;

///////////////////////////////////////////////////////////////////////////////
// Static helper methods
///////////////////////////////////////////////////////////////////////////////

static CCredentials CreateCredentials(
   bool useClientCert,
   const _tstring &storeName,
   const _tstring &certificateName,
   bool useMachineStore,
   DWORD enabledProtocols);

static PCCERT_CONTEXT GetCert(
   const _tstring &storeName,
   const _tstring &certificateName,
   bool useMachineStore);

static void BuildMessagesForConnection(
   CTestMessage::DataLength messageSize,
   CTestMessages &messages);

static void DumpCredentials(
   const CCredentials &credentials);

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   auto result = 2;

   CGlobalErrorHandler errorHandler;

   try
   {
      // Note that maxPendingWrites can be used to limit the number of pending
      // write operations to prevent uncontrolled 'non paged pool' memory usage 
      // if log messages are produced at a rate faster than they can be written 
      // to the disk. Note that once the limit is reached all writes become 
      // synchronous until the number of pending writes has dropped below the 
      // limit again. This is only ever likely to be an issue in a debug build
      // but it WILL affect performance and thread scheduling.

      static const long maxPendingWrites = 100000;

      CPendingWriteLimiter limiter(maxPendingWrites);

      CAsyncFileLog log(_T("SChannelEchoTest.log"), limiter);

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CUsesWinsock usesWinsock;

         CCommandLine commandLine;

         if (commandLine.Parse(argc, argv))
         {
            result = 1;

            const auto messageRepeatCount = commandLine.Messages();

            const auto numConnections = commandLine.Connections();

            const auto transferRate = commandLine.MessageRate();

            const CBufferAllocator::BufferSize messageSize = commandLine.MessageSize();

            const auto delayAfter = commandLine.ConnectionBatchSize();

            const auto delayFor = commandLine.ConnectionBatchPause();

            const auto sendAfterRecv = commandLine.SendAfterRecv();

            const auto syncConnect = commandLine.SyncConnect();

            const auto holdConnections = commandLine.HoldConnections();

            const auto maxSocketsInPool = numConnections;

            const CBufferAllocator::BufferCount maxBuffersInPool = numConnections * 2;

            const CBufferAllocator::BufferSize bufferSize = commandLine.MessageSize();

            const auto fragments = commandLine.Fragments();

            const auto multipleMessages = commandLine.MultipleMessages();

            const auto dataFlowTimeoutMillis = commandLine.MessageFlowTimeout();

            const auto displayConnectionBatching = commandLine.DisplayConnectionBatching();

            CCredentials credentials(
               CreateCredentials(
                  commandLine.UseClientCerts(),
                  commandLine.CertificateStoreName(),
                  commandLine.CertificateName(),
                  commandLine.UseMachineStore(),
                  commandLine.EnabledProtocols()));

            DumpCredentials(credentials);

            CSequencedStreamSocketAllocator socketAllocator(
               maxSocketsInPool,
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               bufferSize,
               maxBuffersInPool);

            CIOPool pool(
               commandLine.NumberOfIOThreads());

            pool.Start();

            CStreamSocketConnectionManager connector(
               credentials,
               commandLine.ReadTimeout(),
               commandLine.IgnoreValidationErrors(),
               pool,
               socketAllocator,
               bufferAllocator);

            const auto addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            if (!connector.AsyncConnectAvailable(address.Type()) && !syncConnect)
            {
               commandLine.ShowHelp(
                  _T("ConnectEx() is not available on this platform.\n")
                  _T("Use -syncConnect to force synchronous connections"));

               return 1;
            }

            if (commandLine.PreAllocate())
            {
               socketAllocator.EnlargePool(maxSocketsInPool);

               bufferAllocator.EnlargePool(maxBuffersInPool);
            }

            // now loop opening connections and kicking things off...

            CMessageCollection messages(numConnections);

            try
            {
               // build messages here and pass in to each connection

               if (messageRepeatCount != 0)
               {
                  cout << "Building Messages" << endl;
               }

               size_t i = 0;

               const CBufferAllocator::BufferSize skipSignOnMessageBytes = commandLine.SkipSignOnMessageBytes();

               for (i = 0; i < numConnections; ++i)
               {
                  auto pMessages = new CTestMessages(
                     i,
                     messageRepeatCount,
                     transferRate,
                     messages,
                     sendAfterRecv,
                     holdConnections,
                     skipSignOnMessageBytes,
                     fragments,
                     multipleMessages,
                     CTestMessages::TrackPendingBytes);

                  BuildMessagesForConnection(messageSize, *pMessages);

                  messages.Add(pMessages);
               }

               cout << "Creating " << numConnections << " connections" << endl;

               i = 0;

               CPerformanceCounter counter(CPerformanceCounter::LockThreadToOneCPU);

               auto ok = true;

               const bool useTargetName = commandLine.UseTargetName();

               const _tstring targetName = commandLine.TargetName();

               const RemoteCredentialValidation remoteCredentialValidation = commandLine.RemoteCredentialValidation();

               for (auto it = messages.Begin(); ok && it != messages.End(); ++it)
               {
                  auto pMessages = *it;

                  if (useTargetName)
                  {
                     if (syncConnect)
                     {
                        CSmartStreamSocket socket = connector.SecureConnectNoThrow(
                           targetName,
                           address,
                           pMessages,
                           remoteCredentialValidation);

                        ok = (socket.Get() != nullptr);
                     }
                     else
                     {
                        CSmartStreamSocket socket = connector.SecureAsyncConnectNoThrow(
                           targetName,
                           address,
                           pMessages,
                           remoteCredentialValidation);

                        ok = (socket.Get() != nullptr);
                     }
                  }
                  else
                  {
                     if (syncConnect)
                     {
                        auto socket = connector.SecureConnectNoThrow(
                           address,
                           pMessages,
                           remoteCredentialValidation);

                        ok = (socket.Get() != nullptr);
                     }
                     else
                     {
                        auto socket = connector.SecureAsyncConnectNoThrow(
                           address,
                           pMessages,
                           remoteCredentialValidation);

                        ok = (socket.Get() != nullptr);
                     }

                     if (delayAfter && (++i % delayAfter == 0))
                     {
                        if (displayConnectionBatching)
                        {
                           cout << i << " connections created" << endl;
                        }

                        if (delayFor)
                        {
                           Sleep(delayFor);
                        }
                     }
                  }
               }

               if (ok)
               {
                  cout << "All connections in progress" << endl;
               }
               else
               {
                  SetEvent(messages.GetConnectionsCompleteEvent());
               }

               CSecurityDescriptorAllowAll sd;

               CSecurityAttributes sa(sd);

               CManualResetEvent shutdownEvent(CGlobalName(_T("JetByteToolsServerShutdown")), &sa);

               HANDLE handlesToWaitFor[2];

               handlesToWaitFor[0] = shutdownEvent.GetWaitHandle();
               handlesToWaitFor[1] = messages.GetConnectionsCompleteEvent();

               auto waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, INFINITE);

               if (waitResult == WAIT_OBJECT_0)
               {
                  result = 1;
               }
               else if (waitResult == WAIT_OBJECT_0 + 1)
               {
                  cout << "All connections complete in " << CStringConverter::TtoA(counter.GetElapsedTimeAsString()) << endl;
                  cout << messages.ConnectionsEstablished() << " established. " << messages.ConnectionsFailed() << " failed."<< endl;

                  if (messages.ConnectionsEstablished())
                  {
                     if (messages.ConnectionsFailed() != 0)
                     {
                        result = 1;
                     }

                     if (messageRepeatCount != 0)
                     {
                        cout << "Waiting for data flow to complete" << endl;

                        const auto checkEvery = commandLine.CheckEvery();

                        Milliseconds elapsedTime = 0;

                        auto done = false;

                        while (!done)
                        {
                           handlesToWaitFor[1] = messages.GetMessagesCompleteEvent();

                           const auto timeout = (dataFlowTimeoutMillis == INFINITE || checkEvery == INFINITE ? INFINITE : min(checkEvery, dataFlowTimeoutMillis - elapsedTime));

                           elapsedTime += timeout;

                           const auto waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, timeout);

                           if (waitResult == WAIT_TIMEOUT)
                           {
                              if (elapsedTime >= dataFlowTimeoutMillis)
                              {
                                 cout << "Waiting for data flow to complete timed out after " << CStringConverter::TtoA(counter.GetElapsedTimeAsString()) << endl;

                                 result = 1;
                                 done = true;
                              }
                              else
                              {
                                 cout << "Waiting for " << messages.GetNumberOfConnectionsWithMessagesPending() << " connections to complete" << endl;

                                 CMessageCollection::StringErrors errors;

                                 if (!messages.CheckMessages(errors))
                                 {
                                    if (!errors.empty())
                                    {
                                       cout << "Message errors:" << endl;

                                       for (CMessageCollection::StringErrors::const_iterator it = errors.begin(); it != errors.end(); ++it)
                                       {
                                          cout << ToStringA(it->second) << " - " << CStringConverter::TtoA(it->first) << endl;
                                       }
                                    }
                                 }
                              }
                           }
                           else if (waitResult == WAIT_OBJECT_0)
                           {
                              result = 1;
                              done = true;
                           }
                           else if (waitResult == WAIT_OBJECT_0 + 1)
                           {
                              const auto elapsedTime = counter.GetElapsedTimeAsString();

                              cout << "Data flow complete in " << CStringConverter::TtoA(elapsedTime) << endl;

                              cout << "Average RTT: " << messages.GetAverageRTT() << "ms" << endl;
                              cout << "    Max RTT: " << messages.GetMaxRTT() << "ms" << endl;
                              cout << "    Min RTT: " << messages.GetMinRTT() << "ms" << endl;

                              result = 0;
                              done = true;
                           }
                           else
                           {
                              throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
                           }
                        }
                     }
                     else
                     {
                        result = 0;
                     }
                  }
               }
               else
               {
                  throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
               }

               if (commandLine.HoldConnections() && commandLine.IsInteractive() && messages.ConnectionsEstablished())
               {
                  cout << "Press return to close connections" << endl;

                  char input;

                  cin >> std::noskipws >> input;
               }
            }
            catch(const CException &e)
            {
               cout << "Exception: " << e.GetDetailsA() << endl;

               result = 1;
            }
            catch(...)
            {
               cout << "Unexpected exception" << endl;

               result = 1;
            }

            if (messageRepeatCount != 0)
            {
               CMessageCollection::StringErrors errors;

               if (!messages.CheckMessages(errors))
               {
                  result = 1;
               }

               if (!errors.empty())
               {
                  cout << "Message errors:" << endl;

                  for (CMessageCollection::StringErrors::const_iterator it = errors.begin(); it != errors.end(); ++it)
                  {
                     cout << ToStringA(it->second) << " - " << CStringConverter::TtoA(it->first) << endl;
                  }

                  result = 1;
               }
            }

            cout << "Wait for connector shutdown..." << endl;

            if (!connector.WaitForShutdownToComplete(10000))
            {
               cout << "Connector shutdown timed out!" << endl;

               result = 1;

               connector.ForceShutdown();
            }

            cout << "Wait for IO pool shutdown..." << endl;

            pool.WaitForShutdownToComplete();

            cout << "Waiting for socket allocator to flush..." << endl;

            socketAllocator.ReleaseSockets();

            cout << "Waiting for buffer allocator to flush..." << endl;

            bufferAllocator.Flush();

            if (!connector.DumpErrors())
            {
               result = 1;
            }

            cout << "All done..." << endl;
         }
      }
      catch(...)
      {
         // TODO
      }
   }
   catch(const CException &e)
   {
      cout << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(const char *pMsg)
   {
      cout << pMsg << endl;

      result = 1;
   }
   catch(...)
   {
      cout << "Unexpected exception" << endl;

      result = 1;
   }

   cout << "Test " << (result == 0 ? "Passed" : "Failed") << endl;

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static void BuildMessagesForConnection(
   const CTestMessage::DataLength messageSize,
   CTestMessages &messages)
{
   messages.AddMessage(new CTestMessage(messageSize));
}

///////////////////////////////////////////////////////////////////////////////
// SChannel specific stuff.
//
// Note: The following code just sets up example SSL server credentials, it's
// been put together from code samples from MSDN. You should read and understand
// the SChannel documentation and then write some real code to initialise
// your credentials in the way that you require.
///////////////////////////////////////////////////////////////////////////////

static CCredentials CreateCredentials(
   const bool useClientCert,
   const _tstring &storeName,
   const _tstring &certificateName,
   const bool useMachineStore,
   DWORD enabledProtocols)
{
   if (useClientCert)
   {
      PCCERT_CONTEXT pCertContext = GetCert(storeName, certificateName, useMachineStore);

      if (!pCertContext)
      {
         throw CException(_T("CreateCredentials"), _T("No certificate available"));
      }

      CCredentials::Data data;

      data.cCreds = 1;
      data.paCred = &pCertContext;

      data.dwFlags = SCH_CRED_NO_DEFAULT_CREDS;

      // Note that for TLS 1.2 you need a certificate that is signed with SHA1 and NOT MD5.
      // If you have an MD5 certificate then TLS 1.2 will not connect and will NOT downgrade
      // to TLS 1.1 or something else...

      DWORD dataSize = 0;

      if (::CertGetCertificateContextProperty(
         pCertContext,
         CERT_SIGN_HASH_CNG_ALG_PROP_ID,
         nullptr,
         &dataSize))
      {
         const ByteBuffer buffer(dataSize);

         BYTE *pData = buffer.GetBuffer();

         if (::CertGetCertificateContextProperty(
            pCertContext,
            CERT_SIGN_HASH_CNG_ALG_PROP_ID,
            pData,
            &dataSize))
         {
            const _tstring signHashDetails(CStringConverter::WtoT(reinterpret_cast<wchar_t *>(pData), dataSize / sizeof(wchar_t)));

            if (signHashDetails.find(_T("MD5")) != _tstring::npos)
            {
               if (enabledProtocols == 0 || enabledProtocols & SP_PROT_TLS1_2)
               {
                  OutputEx(_T("Certificate is signed with: ") + signHashDetails);
                  OutputEx(_T("MD5 hashes do not work with TLS 1.2 and connection downgrade does NOT occur"));
                  OutputEx(_T("This results in failure to connect. Disabling TLS 1.2"));

                  if (enabledProtocols)
                  {
                     enabledProtocols &= ~SP_PROT_TLS1_2;
                  }

                  if (!enabledProtocols)
                  {
                     // if enabledProtocols is zero then all protocols are
                     // enabled, so we need to explicitly enable everything
                     // except TLS1.2

                     enabledProtocols = SP_PROT_SSL3TLS1;
                  }
               }
            }
         }
      }

      data.grbitEnabledProtocols = enabledProtocols;

      CCredentials credentials(data, SECPKG_CRED_OUTBOUND);

      if(pCertContext)
      {
         CertFreeCertificateContext(pCertContext);
      }

      return credentials;
   }
   else
   {
      CCredentials::Data data;

      data.dwFlags = SCH_CRED_NO_DEFAULT_CREDS;

      // Note that for TLS 1.2 you need a certificate that is signed with SHA1 and NOT MD5.
      // If you have an MD5 certificate then TLS 1.2 will not connect and will NOT downgrade
      // to TLS 1.1 or something else...

      data.grbitEnabledProtocols = enabledProtocols;

      CCredentials credentials(data, SECPKG_CRED_OUTBOUND);

      return credentials;
   }
}

#pragma comment(lib, "Crypt32.lib")

static PCCERT_CONTEXT GetCert(
   const _tstring &storeName,
   const _tstring &certificateName,
   const bool useMachineStore)
{
   if (storeName.length() == 0)
   {
      throw CException(_T("GetCert()"), _T("No certificate store name specified!"));
   }

   if (certificateName.length() == 0)
   {
      throw CException(_T("GetCert()"), _T("No certificate name specified!"));
   }

   HCERTSTORE hMyCertStore ;

#ifdef _UNICODE
#define JETBYTE_CERT_STORE_PROV_SYSTEM CERT_STORE_PROV_SYSTEM_W
#else
#define JETBYTE_CERT_STORE_PROV_SYSTEM CERT_STORE_PROV_SYSTEM_A
#endif

   if (useMachineStore)
   {
      hMyCertStore = CertOpenStore(
         JETBYTE_CERT_STORE_PROV_SYSTEM,
         X509_ASN_ENCODING,
         0,
         CERT_SYSTEM_STORE_LOCAL_MACHINE,
         storeName.c_str());
   }
   else
   {
      hMyCertStore = CertOpenSystemStore(0, storeName.c_str());
   }

   if (!hMyCertStore)
   {
      const DWORD lastError = GetLastError();

      OutputEx(_T("CertOpenSystemStore() - failed - ") + ErrorCodeToErrorMessage(lastError));

      return nullptr;
   }

#ifdef _UNICODE
#define JETBYTE_CERT_FIND_SUBJECT_STR CERT_FIND_SUBJECT_STR_W
#else
#define JETBYTE_CERT_FIND_SUBJECT_STR CERT_FIND_SUBJECT_STR_A
#endif

    // Find certificate. Note that this sample just searchs for a
    // certificate that contains the user name somewhere in the subject name.
    // A real application should be a bit less casual.
   const PCCERT_CONTEXT pCertContext = CertFindCertificateInStore(
      hMyCertStore,
      X509_ASN_ENCODING,
      0,
      JETBYTE_CERT_FIND_SUBJECT_STR,
      certificateName.c_str(),
      nullptr);

   if (!pCertContext)
   {
      const DWORD lastError = GetLastError();

      OutputEx(_T("CertFindCertificateInStore() - \"") + certificateName + _T("\" - ") + ErrorCodeToErrorMessage(lastError));
   }

   return pCertContext;
}

static void DumpCredentials(
   const CCredentials &credentials)
{
   const ByteBuffer buffer(4096);

   CredHandle hCred = credentials;

   SECURITY_STATUS status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_SUPPORTED_ALGS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query supported algorithms"), status);

   const auto *pAlgs = reinterpret_cast<const _SecPkgCred_SupportedAlgs *>(buffer.GetBuffer());

   for (DWORD i = 0; i < pAlgs->cSupportedAlgs; ++i)
   {
      const _tstring asString = CContext::GetAlgorithmAsString(pAlgs->palgSupportedAlgs[i]);

      if (asString.length())
      {
         OutputEx(ToHexString(pAlgs->palgSupportedAlgs[i]) + _T(" - ") + asString);
      }
      else
      {
         OutputEx(ToHexString(pAlgs->palgSupportedAlgs[i]));
      }
   }

   status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_CIPHER_STRENGTHS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query cipher strengths"), status);

   const auto *pCipher = reinterpret_cast<const _SecPkgCred_CipherStrengths *>(buffer.GetBuffer());

   OutputEx(_T("Min cipher strength: ") + ToString(pCipher->dwMinimumCipherStrength));
   OutputEx(_T("Max cipher strength: ") + ToString(pCipher->dwMaximumCipherStrength));

   status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_SUPPORTED_PROTOCOLS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query supported protocols"), status);

   const auto *pProto = reinterpret_cast<const _SecPkgCred_SupportedProtocols  *>(buffer.GetBuffer());

   OutputEx(JetByteTools::Core::ToHexString(pProto->grbitProtocol) + _T(" - ") + CContext::GetProtocolAsString(pProto->grbitProtocol));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
