///////////////////////////////////////////////////////////////////////////////
// File: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "CommandLine.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/Tchar.h"

#include "JetByteTools/Win32Tools/SystemUtils.h"

#pragma hdrstop

#include <iostream>

#include <schannel.h>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::_tstring;

using JetByteTools::Win32::GetComputerName;

using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Constants
///////////////////////////////////////////////////////////////////////////////

static const _tstring s_requireClientCertsSwitch                     = _T("requireClientCerts");
static const _tstring s_allowUntrustedRootSwitch      = _T("allowUntrustedRoot");
static const _tstring s_ignoreValidationErrorsSwitch  = _T("ignoreValidationErrors");
static const _tstring s_useMachineStoreSwitch                        = _T("useMachineStore");
static const _tstring s_messageSizeArgument                          = _T("messageSize");
static const _tstring s_initialBusinessLogicThreadsArgument          = _T("initialBusinessLogicThreads");
static const _tstring s_minimumBusinessLogicThreadsArgument          = _T("minimumBusinessLogicThreads");
static const _tstring s_maximumBusinessLogicThreadsArgument          = _T("maximumBusinessLogicThreads");
static const _tstring s_maximumDormantThreadsArgument                = _T("maximumDormantThreads");
static const _tstring s_businessLogicPoolMaintenancePeriodArgument   = _T("businessLogicPoolMaintenancePeriod");
static const _tstring s_businessLogicPoolDispatchTimeoutArgument     = _T("businessLogicPoolDispatchTimeout");
static const _tstring s_storeNameArgument                            = _T("storeName");
static const _tstring s_certificateNameArgument                      = _T("certificateName");
static const _tstring s_useDefaultsSwitch                            = _T("useDefaults");
static const _tstring s_useSSL2Switch                 = _T("ssl2");
static const _tstring s_useSSL3Switch                 = _T("ssl3");
static const _tstring s_useTLS10Switch                = _T("tls1");
static const _tstring s_useTLS11Switch                = _T("tls1.1");
static const _tstring s_useTLS12Switch                = _T("tls1.2");

///////////////////////////////////////////////////////////////////////////////
// CSChannelServerCommandLine
///////////////////////////////////////////////////////////////////////////////

CSChannelServerCommandLine::CSChannelServerCommandLine(
   const DWORD argc,
   const char **argv)
   :  CCommandLine(
         argc,
         argv,
         _T("SChannelThreadPoolLargePacketEchoServer"),
         CCommandLine::TCPServer |
         CCommandLine::SupportsSendRecvBufSize |
         CCommandLine::SupportsZeroByteRead)
{
   m_parser.AddSwitch(s_requireClientCertsSwitch);
   m_parser.AddSwitch(s_allowUntrustedRootSwitch);
   m_parser.AddSwitch(s_ignoreValidationErrorsSwitch);
   m_parser.AddSwitch(s_useMachineStoreSwitch);
   m_parser.AddSwitch(s_useDefaultsSwitch);
   m_parser.AddSwitch(s_useSSL2Switch);
   m_parser.AddSwitch(s_useSSL3Switch);
   m_parser.AddSwitch(s_useTLS10Switch);
   m_parser.AddSwitch(s_useTLS11Switch);
   m_parser.AddSwitch(s_useTLS12Switch);

   m_parser.AddArgument(s_storeNameArgument);
   m_parser.AddArgument(s_certificateNameArgument);

   m_parser.AddArgument(s_messageSizeArgument);
   m_parser.AddArgument(s_initialBusinessLogicThreadsArgument);
   m_parser.AddArgument(s_minimumBusinessLogicThreadsArgument);
   m_parser.AddArgument(s_maximumBusinessLogicThreadsArgument);
   m_parser.AddArgument(s_maximumDormantThreadsArgument);
   m_parser.AddArgument(s_businessLogicPoolMaintenancePeriodArgument);
   m_parser.AddArgument(s_businessLogicPoolDispatchTimeoutArgument);
}

bool CSChannelServerCommandLine::Parse()
{
   bool ok = CCommandLine::Parse();

   if (ok)
   {
      _tstring message;

      if (m_parser.IsSet(s_allowUntrustedRootSwitch) &&
          m_parser.IsSet(s_ignoreValidationErrorsSwitch))
      {
         ok = false;

         message = _T("only one of -allowUntrustedRoot is implied by -ignoreValidationErrors!");
      }

      if (!ok)
      {
         ShowHelp(message);
      }
   }

   return ok ;
}

CCommandLine::Count CSChannelServerCommandLine::MessageSize() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_messageSizeArgument, _T("1024")).c_str());
}

bool CSChannelServerCommandLine::RequireClientCerts() const
{
   return m_parser.IsSet(s_requireClientCertsSwitch);
}

CCommandLine::Count CSChannelServerCommandLine::InitialBusinessLogicThreads() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_initialBusinessLogicThreadsArgument, _T("5")).c_str());
}

CCommandLine::Count CSChannelServerCommandLine::MinimumBusinessLogicThreads() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_minimumBusinessLogicThreadsArgument, _T("5")).c_str());
}

CCommandLine::Count CSChannelServerCommandLine::MaximumBusinessLogicThreads() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_maximumBusinessLogicThreadsArgument, _T("20")).c_str());
}

CCommandLine::Count CSChannelServerCommandLine::MaximumDormantThreads() const
{
   return StringToCount(m_parser.GetOptionalValueAsString(s_maximumDormantThreadsArgument, _T("5")).c_str());
}

Milliseconds CSChannelServerCommandLine::BusinessLogicPoolMaintenancePeriod() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_businessLogicPoolMaintenancePeriodArgument, _T("5000")).c_str());
}

Milliseconds CSChannelServerCommandLine::BusinessLogicPoolDispatchTimeout() const
{
   return _ttol(m_parser.GetOptionalValueAsString(s_businessLogicPoolDispatchTimeoutArgument, _T("100")).c_str());
}

bool CSChannelServerCommandLine::AllowUntrustedRootSwitch() const
{
   return m_parser.IsSet(s_allowUntrustedRootSwitch);
}

bool CSChannelServerCommandLine::IgnoreValidationErrors() const
{
   return m_parser.IsSet(s_ignoreValidationErrorsSwitch);
}

bool CSChannelServerCommandLine::UseMachineStore() const
{
   return m_parser.IsSet(s_useMachineStoreSwitch);
}

_tstring CSChannelServerCommandLine::CertificateStoreName() const
{
   return m_parser.GetOptionalValueAsString(s_storeNameArgument, _T("MY"));
}

_tstring CSChannelServerCommandLine::CertificateName() const
{
   return m_parser.GetOptionalValueAsString(s_certificateNameArgument, GetComputerName());
}

bool CSChannelServerCommandLine::UseDefaultsToLocateCertificate() const
{
   return m_parser.IsSet(s_useDefaultsSwitch);
}

DWORD CSChannelServerCommandLine::EnabledProtocols() const
{
   DWORD enabledProtocols = 0;      // default to all!

   if (m_parser.IsSet(s_useSSL2Switch))
   {
      enabledProtocols |= SP_PROT_SSL2;
   }

   if (m_parser.IsSet(s_useSSL3Switch))
   {
      enabledProtocols |= SP_PROT_SSL3;
   }

   if (m_parser.IsSet(s_useTLS10Switch))
   {
      enabledProtocols |= SP_PROT_TLS1_0;
   }

   if (m_parser.IsSet(s_useTLS11Switch))
   {
      enabledProtocols |= SP_PROT_TLS1_1;
   }

   if (m_parser.IsSet(s_useTLS12Switch))
   {
      enabledProtocols |= SP_PROT_TLS1_2;
   }

   return enabledProtocols;
}

void CSChannelServerCommandLine::ShowMoreHelp()
{
   cout << " o -requireClientCerts      Requires a client to have a client cert." << endl;
   cout << " o -allowUntrustedRoot      Allows self signed certificates." << endl;
   cout << " o -ignoreValidationErrors  Ignores all certificate validation errors." << endl;
   cout << " o -useMachineStore" << endl;
   cout << " o -storeName               Defaults to MY." << endl;
   cout << " o -certificateName         Defaults to the name of the computer" << endl;
   cout << " o -useDefaults             If the specified certificate isn't found, then" << endl;
   cout << "                            also try the machine name and the user name in" << endl;
   cout << "                            the MY store." << endl;
   cout << endl;
   cout << " Enabled protocols" << endl;
   cout << " By default all supported protocols are enabled. If you specify any of the" << endl;
   cout << " command line switches below then ONLY the protocols specified will be" << endl;
   cout << " enabled. So if you want JUST TLS1.1 and TLS1.2 you need to specify" << endl;
   cout << " -tls1.1 -tls1.2 on the command line." << endl;
   cout << endl;
   cout << " o -ssl2                    Enable SSL 2" << endl;
   cout << " o -ssl2                    Enable SSL 3" << endl;
   cout << " o -tls1                    Enable TLS 1.0" << endl;
   cout << " o -tls1.1                  Enable TLS 1.1" << endl;
   cout << " o -tls1.2                  Enable TLS 1.2" << endl;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: CommandLine.cpp
///////////////////////////////////////////////////////////////////////////////
