///////////////////////////////////////////////////////////////////////////////
// File: MessageAccumulator.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2016 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "PerConnectionData.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/IOTools/BufferBasedBufferAllocator.h"
#include "JetByteTools/IOTools/CustomPoolSmartBuffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#pragma hdrstop

#include <algorithm>    // for min

#ifdef DEBUG_ECHO_STYLE
#define DEBUGGING(_m) JetByteTools::Core::Output(_m)
#else
#define DEBUGGING(_m)
#endif

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::GetStringLength;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::CLockableObject;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::CBufferChain;
using JetByteTools::IO::CBufferBasedBufferAllocator;
using JetByteTools::IO::CCustomPoolSmartBuffer;

using JetByteTools::Socket::IStreamSocket;

using std::min;

///////////////////////////////////////////////////////////////////////////////
// CPerConnectionData
///////////////////////////////////////////////////////////////////////////////

CPerConnectionData::CPerConnectionData(
   const IIndexedOpaqueUserData::UserDataIndex bufferChainIndex,
   const IBuffer::BufferSize maxMessageLength,
   const bool useMultipleWrites)
   :  CMessageAccumulator(maxMessageLength),
      m_outOfSequenceReads(m_lock, bufferChainIndex),
      m_useMultipleWrites(useMultipleWrites),
      m_writeAsBytes(false),//true),
      m_processHow(ProcessContiguously),
      m_processContiguouslyThisTime(true),
      m_useMultipleWritesThisTime(false),
      m_writeAsBytesThisTime(false)
{
}

void CPerConnectionData::ProcessMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer_)
{
   CCustomPoolSmartBuffer buffer(m_outOfSequenceReads.GetNext(buffer_), socket);

   while (buffer.Get())
   {
      // Reconstitute the buffer chain that makes up the complete message...

      CBufferChain buffers;

      buffers.Attach(buffer.Detach());

      if (m_processHow == ProcessContiguously ||
         (m_processHow == ProcessAlternate && m_processContiguouslyThisTime))
      {
         DEBUGGING(_T("Echo contiguosly"));

         // if we need to process the data as one contiguous block of memory then we need to
         // do this...

         CSmartBuffer contiguousMessage(buffers.RemoveAsSingleBuffer());

         InternalProcessMessage(socket, contiguousMessage);

         m_processContiguouslyThisTime = false;
      }
      else
      {
         // if we can process it as discrete buffers of arbitrary size then we can do this

         DEBUGGING(_T("Echo chain"));

         InternalProcessMessage(socket, buffers);

         m_processContiguouslyThisTime = true;
      }

      buffer = m_outOfSequenceReads.ProcessAndGetNext();
   }
}

void CPerConnectionData::InternalProcessMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   if (m_useMultipleWrites && m_useMultipleWritesThisTime)
   {
      DEBUGGING(_T("Echo with multiple writes"));

      CLockableObject::Owner lock(m_lock);

      IBuffer::BufferSize bytesSent = 0;

      const BYTE *pMessageData = buffer->GetMemory();
      const IBuffer::BufferSize messageDataLength = buffer->GetUsed();

      while (bytesSent < messageDataLength)
      {
         const IBuffer::BufferSize responsePacketSize = 666;

         const IBuffer::BufferSize bytesToSend = min(responsePacketSize, messageDataLength - bytesSent);

         socket.TryWrite(pMessageData + bytesSent, bytesToSend);

         bytesSent += bytesToSend;
      }

      m_useMultipleWritesThisTime = false;
   }
   else
   {
      if (m_writeAsBytes && m_writeAsBytesThisTime)
      {
         DEBUGGING(_T("Echo as bytes"));

         const BYTE *pData = buffer->GetMemory();

         const IBuffer::BufferSize dataLength = buffer->GetUsed();

         socket.TryWrite(pData, dataLength);

         m_writeAsBytesThisTime = false;
      }
      else
      {
         DEBUGGING(_T("Echo as single write"));

         // Echo packet with a single write.

         socket.TryWrite(buffer);

         m_writeAsBytesThisTime = true;
      }

      m_useMultipleWritesThisTime = true;
   }
}

void CPerConnectionData::InternalProcessMessage(
   IStreamSocket &socket,
   CBufferChain &buffers)
{
   if (m_useMultipleWrites && m_useMultipleWritesThisTime)
   {
      DEBUGGING(_T("Echo chain as multiple writes"));

      CLockableObject::Owner lock(m_lock);

      CSmartBuffer buffer(buffers.GetNext());

      while (buffer.Get())
      {
         IBuffer::BufferSize bytesSent = 0;

         const BYTE *pMessageData = buffer->GetMemory();
         const IBuffer::BufferSize messageDataLength = buffer->GetUsed();

         while (bytesSent < messageDataLength)
         {
            const IBuffer::BufferSize responsePacketSize = 666;

            const IBuffer::BufferSize bytesToSend = min(responsePacketSize, messageDataLength - bytesSent);

            socket.TryWrite(pMessageData + bytesSent, bytesToSend);

            bytesSent += bytesToSend;
         }

         buffer = buffers.GetNext();
      }

      m_useMultipleWritesThisTime = false;
   }
   else
   {
      DEBUGGING(_T("Echo chain as single write"));

      // Echo packet with a single write.
      socket.TryWrite(buffers);

      m_useMultipleWritesThisTime = true;
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: PerConnectionData.cpp
///////////////////////////////////////////////////////////////////////////////
