///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "SocketServer.h"
#include "ThreadPool.h"
#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"
#include "JetByteTools/CoreTools/ZeroInitialiseExpandableBuffer.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/Win32Tools/SystemUtils.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"
#include "JetByteTools/SocketTools/FullAddress.h"

#include "JetByteTools/SSPITools/Exception.h"

#include "JetByteTools/SChannelTools/Context.h"
#include "JetByteTools/SChannelTools/Credentials.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::ToHexString;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CGlobalErrorHandler;
using JetByteTools::Core::ErrorCodeToErrorMessage;

using JetByteTools::Win32::GetComputerName;
using JetByteTools::Win32::GetUserName;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::ZeroByteReadConfiguration;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;

using JetByteTools::SSPI::SChannel::CCredentials;
using JetByteTools::SSPI::SChannel::CContext;

using JetByteTools::IO::CAsyncFileLog;

typedef JetByteTools::Core::TZeroInitialiseExpandableBuffer<BYTE> ByteBuffer;

using std::cerr;
using std::endl;
using std::string;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions, SChannel specific...
///////////////////////////////////////////////////////////////////////////////

static CCredentials CreateCredentials(
   const _tstring &storeName,
   const _tstring &certificateName,
   bool useMachineStore,
   bool attemptDefaults,
   DWORD enabledProtocols);

static PCCERT_CONTEXT GetCert(
   const _tstring &storeName,
   const _tstring &certificateName,
   bool useMachineStore);

static void DumpCredentials(
   const CCredentials &credentials);

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char *argv[ ])
{
   CGlobalErrorHandler errorHandler;

   int result = 0;

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("SChannelThreadPoolLargePacketEchoServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CSChannelServerCommandLine commandLine(argc, argv);

         if (commandLine.Parse())
         {
            CCredentials credentials(
               CreateCredentials(
                  commandLine.CertificateStoreName(),
                  commandLine.CertificateName(),
                  commandLine.UseMachineStore(),
                  commandLine.UseDefaultsToLocateCertificate(),
                  commandLine.EnabledProtocols()));

            DumpCredentials(credentials);

            CIOPool ioPool(
               commandLine.NumberOfIOThreads(),
               commandLine.DisplayDebug());

            ioPool.Start();

            CSequencedStreamSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            CThreadPool businessLogicPool(
               commandLine.InitialBusinessLogicThreads(),
               commandLine.MinimumBusinessLogicThreads(),
               commandLine.MaximumBusinessLogicThreads(),
               commandLine.MaximumDormantThreads(),
               commandLine.BusinessLogicPoolMaintenancePeriod(),
               commandLine.BusinessLogicPoolDispatchTimeout());

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            const bool useMultipleWrites = false;

            CSocketServer server(
               businessLogicPool,
               65536,                        // max message size
               useMultipleWrites,
               credentials,
               commandLine.RequireClientCerts(),
               commandLine.AllowUntrustedRootSwitch(),
               commandLine.IgnoreValidationErrors(),
               address,
               commandLine.ListenBacklog(),
               ioPool,
               socketAllocator,
               bufferAllocator,
               commandLine.RecvBufferSize(),
               commandLine.SendBufferSize(),
               commandLine.GetZeroByteReadConfiguration());

            businessLogicPool.Start(server);

            server.Start();

            server.StartAcceptingConnections();

            shutdownHandler.SetServerControl(server, commandLine.EventNamePrefix());

            shutdownHandler.WaitForShutdownRequest();

            server.WaitForShutdownToComplete();

            businessLogicPool.WaitForShutdownToComplete(15000);

            ioPool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// SChannel specific stuff.
//
// Note: The following code just sets up example SSL server credentials, it's
// been put together from code samples from MSDN. You should read and understand
// the SChannel documentation and then write some real code to initialise
// your credentials in the way that you require.
///////////////////////////////////////////////////////////////////////////////

static CCredentials CreateCredentials(
   const _tstring &storeName,
   const _tstring &certificateName,
   const bool useMachineStore,
   const bool attemptDefaults,
   DWORD enabledProtocols)
{
   PCCERT_CONTEXT pCertContext = GetCert(storeName, certificateName, useMachineStore);

   if (!pCertContext && attemptDefaults)
   {
      const _tstring computerName = GetComputerName();

      if (computerName != certificateName || storeName != _T("MY"))
      {
         pCertContext = GetCert(storeName, computerName, false);

         if (!pCertContext)
         {
            pCertContext = GetCert(storeName, computerName, true);
         }
      }

      if (!pCertContext)
      {
         const _tstring userName = GetUserName();

         if (userName != certificateName || storeName != _T("MY"))
         {
            pCertContext = GetCert(storeName, userName, false);

            if (!pCertContext)
            {
               pCertContext = GetCert(storeName, userName, true);
            }
         }
      }
   }

   if (!pCertContext)
   {
      throw CException(_T("CreateCredentials()"), _T("No certificate available"));
   }

   CCredentials::Data data;

   data.cCreds = 1;
   data.paCred = &pCertContext;

   // Note that for TLS 1.2 you need a certificate that is signed with SHA1 and NOT MD5.
   // If you have an MD5 certificate then TLS 1.2 will not connect and will NOT downgrade
   // to TLS 1.1 or something else...

   DWORD dataSize = 0;

   if (::CertGetCertificateContextProperty(
      pCertContext,
      CERT_SIGN_HASH_CNG_ALG_PROP_ID,
      nullptr,
      &dataSize))
   {
      const ByteBuffer buffer(dataSize);

      BYTE *pData = buffer.GetBuffer();

      if (::CertGetCertificateContextProperty(
         pCertContext,
         CERT_SIGN_HASH_CNG_ALG_PROP_ID,
         pData,
         &dataSize))
      {
         const _tstring signHashDetails(CStringConverter::WtoT(reinterpret_cast<wchar_t *>(pData), dataSize / sizeof(wchar_t)));

         OutputEx(_T("Certificate is signed with: ") + signHashDetails);

         if (signHashDetails.find(_T("MD5")) != _tstring::npos)
         {
            if (enabledProtocols == 0 || enabledProtocols & SP_PROT_TLS1_2)
            {
               OutputEx(_T("Certificate is signed with: ") + signHashDetails);
               OutputEx(_T("MD5 hashes do not work with TLS 1.2 and connection downgrade does NOT occur"));
               OutputEx(_T("This results in failure to connect. Disabling TLS 1.2"));

               if (enabledProtocols)
               {
                  enabledProtocols &= ~SP_PROT_TLS1_2;
               }

               if (!enabledProtocols)
               {
                  // if enabledProtocols is zero then all protocols are
                  // enabled, so we need to explicitly enable everything
                  // except TLS1.2

                  enabledProtocols = SP_PROT_SSL3TLS1;
               }
            }
         }
      }
   }

   data.grbitEnabledProtocols = enabledProtocols;

   CCredentials credentials(data, SECPKG_CRED_INBOUND);

   if(pCertContext)
   {
      CertFreeCertificateContext(pCertContext);
   }

   return credentials;
}

#pragma comment(lib, "Crypt32.lib")

static PCCERT_CONTEXT GetCert(
   const _tstring &storeName,
   const _tstring &certificateName,
   const bool useMachineStore)
{
   if (storeName.length() == 0)
   {
      throw CException(_T("GetCert()"), _T("No certificate store name specified!"));
   }

   if (certificateName.length() == 0)
   {
      throw CException(_T("GetCert()"), _T("No certificate name specified!"));
   }

   HCERTSTORE hMyCertStore ;

#ifdef _UNICODE
#define JETBYTE_CERT_STORE_PROV_SYSTEM CERT_STORE_PROV_SYSTEM_W
#else
#define JETBYTE_CERT_STORE_PROV_SYSTEM CERT_STORE_PROV_SYSTEM_A
#endif

   if (useMachineStore)
   {
      hMyCertStore = CertOpenStore(
         JETBYTE_CERT_STORE_PROV_SYSTEM,
         X509_ASN_ENCODING,
         0,
         CERT_SYSTEM_STORE_LOCAL_MACHINE,
         storeName.c_str());
   }
   else
   {
      hMyCertStore = CertOpenSystemStore(0, storeName.c_str());
   }

   if (!hMyCertStore)
   {
      const DWORD lastError = GetLastError();

      OutputEx(_T("CertOpenSystemStore() - failed - ") + ErrorCodeToErrorMessage(lastError));

      return nullptr;
   }

#ifdef _UNICODE
#define JETBYTE_CERT_FIND_SUBJECT_STR CERT_FIND_SUBJECT_STR_W
#else
#define JETBYTE_CERT_FIND_SUBJECT_STR CERT_FIND_SUBJECT_STR_A
#endif

    // Find certificate. Note that this sample just searchs for a
    // certificate that contains the user name somewhere in the subject name.
    // A real application should be a bit less casual.
   const PCCERT_CONTEXT pCertContext = CertFindCertificateInStore(
      hMyCertStore,
      X509_ASN_ENCODING,
      0,
      JETBYTE_CERT_FIND_SUBJECT_STR,
      certificateName.c_str(),
      NULL);

   if (!pCertContext)
   {
      const DWORD lastError = GetLastError();

      OutputEx(_T("CertFindCertificateInStore() - \"") + certificateName + _T("\" - ") + ErrorCodeToErrorMessage(lastError));
   }

   return pCertContext;
}

static void DumpCredentials(
   const CCredentials &credentials)
{
   const ByteBuffer buffer(4096);

   CredHandle hCred = credentials;

   SECURITY_STATUS status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_SUPPORTED_ALGS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query supported algorithms"), status);

   const _SecPkgCred_SupportedAlgs *pAlgs = reinterpret_cast<const _SecPkgCred_SupportedAlgs *>(buffer.GetBuffer());

   for (DWORD i = 0; i < pAlgs->cSupportedAlgs; ++i)
   {
      const _tstring asString = CContext::GetAlgorithmAsString(pAlgs->palgSupportedAlgs[i]);

      if (asString.length())
      {
         OutputEx(ToHexString(pAlgs->palgSupportedAlgs[i]) + _T(" - ") + asString);
      }
      else
      {
         OutputEx(ToHexString(pAlgs->palgSupportedAlgs[i]));
      }
   }

   status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_CIPHER_STRENGTHS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query cipher strengths"), status);

   const _SecPkgCred_CipherStrengths *pCipher = reinterpret_cast<const _SecPkgCred_CipherStrengths *>(buffer.GetBuffer());

   OutputEx(_T("Min cipher strength: ") + ToString(pCipher->dwMinimumCipherStrength));
   OutputEx(_T("Max cipher strength: ") + ToString(pCipher->dwMaximumCipherStrength));

   status = ::QueryCredentialsAttributes(&hCred, SECPKG_ATTR_SUPPORTED_PROTOCOLS, buffer.GetBuffer());

   JetByteTools::SSPI::CException::ThrowOnFailure(_T("DumpCredentials() - Query supported protocols"), status);

   const _SecPkgCred_SupportedProtocols  *pProto = reinterpret_cast<const _SecPkgCred_SupportedProtocols  *>(buffer.GetBuffer());

   OutputEx(JetByteTools::Core::ToHexString(pProto->grbitProtocol) + _T(" - ") + CContext::GetProtocolAsString(pProto->grbitProtocol));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
