///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2008 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"
#include "PerConnectionData.h"

#include "JetByteTools/IOTools/IBuffer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#include "JetByteTools/Win32Tools/GetLastErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#include "JetByteTools/IOTools/SmartBuffer.h"
#include "JetByteTools/IOTools/BufferChain.h"

#include "JetByteTools/SSPITools/Exception.h"

#pragma hdrstop

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToHexString;

using JetByteTools::Win32::GetLastErrorMessage;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::CBufferChain;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::SocketBufferSize;
using JetByteTools::Socket::ZeroByteReadConfiguration;

using JetByteTools::SSPI::SChannel::CCredentials;
using JetByteTools::SSPI::SChannel::CContext;

using std::min;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static void DisplayCertChain(
   const CERT_CONTEXT *pClientCert);

static HRESULT VerifyClientCertificate(
   const CERT_CONTEXT *pClientCert);

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   CThreadPool &businessLogicPool,
   const IBuffer::BufferSize maxMessageLength,
   const bool useMultipleWrites,
   CCredentials &credentials,
   const bool verifyPeer,
   const bool allowUntrustedRoot,
   const bool ignoreValidationErrors,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   const ZeroByteReadConfiguration zeroByteReadConfiguration)
   :  Base(
         credentials,
         verifyPeer,
         address,
         listenBacklog,
         *this,
         pool,
         socketAllocator,
         bufferAllocator,
         recvBufferSize,
         sendBufferSize,
         zeroByteReadConfiguration),
      m_maxMessageLength(maxMessageLength),
      m_useMultipleWrites(useMultipleWrites),
      m_pool(businessLogicPool),
      m_perConnectionIndex(socketAllocator.RequestUserDataSlot(_T("CSocketServer::m_perConnectionIndex"))),
      m_bufferChainIndex(bufferAllocator.RequestUserDataSlot(_T("CSocketServer::BufferChain"))),
      m_allowUntrustedRoot(allowUntrustedRoot),
      m_ignoreValidationErrors(ignoreValidationErrors)
{

}

CSocketServer::~CSocketServer()
{
   Base::WaitForShutdownToComplete();
}

void CSocketServer::OnSecureConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnSecureConnectionEstablished")));

   // Associate our user data with the socket

   CPerConnectionData *pPerConnectionData = new CPerConnectionData(
      m_bufferChainIndex,
      m_maxMessageLength,
      m_useMultipleWrites);

   socket.SetUserPointer(m_perConnectionIndex, pPerConnectionData);
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   DEBUG_ONLY(Output(_T("OnConnectionEstablished")));

   socket.TryRead();
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
   DEBUG_ONLY(Output(_T("OnSocketReleased")));
}

void CSocketServer::OnSecureSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   DEBUG_ONLY(Output(_T("OnSecureSocketReleased")));

   CPerConnectionData *pPerConnectionData = static_cast<CPerConnectionData*>(userData.GetUserPointer(m_perConnectionIndex));

   userData.SetUserPointer(m_perConnectionIndex, nullptr);

   delete pPerConnectionData;
}

void CSocketServer::OnError(
   const _tstring &message)
{
   OutputEx(_T("OnError - ") + message);
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   ProcessDataStream(socket, buffer);

   socket.TryRead();
}

void CSocketServer::ProcessDataStream(
   IStreamSocket &socket,
   CSmartBuffer &buffer) const
{
   CPerConnectionData *pPerConnectionData = static_cast<CPerConnectionData*>(socket.GetUserPointer(m_perConnectionIndex));

   do
   {
      // Accumulate our message until we have a complete one and then process it...

      CSmartBuffer message(pPerConnectionData->Add(buffer));

      if (message.Get())
      {
         // message complete...

         m_pool.DispatchCompleteMessage(socket, message);
      }
   }
   while (buffer.Get());
}

void CSocketServer::ProcessMessageOnWorkerThread(
   IStreamSocket &socket,
   CSmartBuffer &buffer,
   const bool afterShutdown) const
{
   try
   {
      if (afterShutdown)
      {
         CBufferChain buffers;

         buffers.Attach(buffer.Detach());
      }
      else
      {
         CPerConnectionData *pPerConnectionData = static_cast<CPerConnectionData*>(socket.GetUserPointer(m_perConnectionIndex));

         pPerConnectionData->ProcessMessage(socket, buffer);
      }
   }
   catch(const CException &e)
   {
      OutputEx(_T("Process - Exception - ") + e.GetDetails());

      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("Process - Unexpected exception"));

      socket.AbortConnection();
   }
}

void CSocketServer::FatalError(
   IStreamSocket & /*socket*/,
   const SECURITY_STATUS status)
{
   Output(_T("CSocketServer::FatalError() - ") + JetByteTools::SSPI::CException::GetErrorMessage(status));
}

void CSocketServer::OnContextComplete(
   IStreamSocket & /*socket*/,
   const CContext &context)
{
   Output(_T("CSocketServer::OnContextComplete() - Secure connection is established"));

   SecPkgContext_ConnectionInfo info;

   context.QueryContextAttributes(info);

   Output(_T("Connection information"));
   Output(_T("       Attributes : ") + CContext::GetAttributesAsString(context.GetAttributes())  + _T(" [") + ToHexString(context.GetAttributes()) + _T("]"));
   Output(_T("         Protocol : ") + CContext::GetProtocolAsString(info.dwProtocol) + _T(" [") + ToHexString(info.dwProtocol) + _T("]"));
   Output(_T("           Cipher : ") + CContext::GetAlgorithmAsString(info.aiCipher) + _T(" [") + ToHexString(info.aiCipher) + _T("]"));
   Output(_T("  Cipher Strength : ") + ToString(info.dwCipherStrength));
   Output(_T("             Hash : ") + CContext::GetAlgorithmAsString(info.aiHash) + _T(" [") + ToHexString(info.aiHash) + _T("]"));
   Output(_T("    Hash Strength : ") + ToString(info.dwHashStrength));
   Output(_T("         Exchange : ") + CContext::GetAlgorithmAsString(info.aiExch) + _T(" [") + ToHexString(info.aiExch) + _T("]"));
   Output(_T("Exchange Strength : ") + ToString(info.dwExchStrength));
}

bool CSocketServer::Verify(
   IStreamSocket &socket,
   const _tstring & /*notValidForServer*/,
   const CContext &context)
{
   bool ok = true;

   Output(_T("CSocketServer::Verify() - do manual certificate verification here (only called if we REQUIRE client certs)"));

   CERT_CONTEXT *pCertContext = context.GetRemoteCertificate();

   if (pCertContext)
   {
      Output(_T("Verify client certificate here"));

      DisplayCertChain(pCertContext);

      // Attempt to validate client certificate.

      const SECURITY_STATUS status = VerifyClientCertificate(
         pCertContext);

      if (status != SEC_E_OK)
      {
         // The client certificate did not validate correctly. At this
         // point, we cannot tell if we are connecting to the correct
         // client, or if we are connecting to a "man in the middle"
         // attack server.

         // It is therefore best if we abort the connection.

         if (m_ignoreValidationErrors ||
            (status == CERT_E_UNTRUSTEDROOT && m_allowUntrustedRoot))
         {
            Output(_T("Certificate Verification Error IGNORED - ") + GetLastErrorMessage(status));

            ok = true;
         }
         else
         {
            Output(_T("Certificate Verification Error - ") + GetLastErrorMessage(status));

            ok = false;
         }
       }

      ::CertFreeCertificateContext(pCertContext);
   }
   else
   {
      Output(_T("Client has no certificate"));

      if (RequiresPeerVerification(socket))
      {
         Output(_T("Client certificate is required"));

         ok = false;
      }
   }

   return ok;
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static void DisplayCertChain(
   const CERT_CONTEXT *pClientCert)
{
   if (pClientCert)
   {
      TCHAR szName[1000];

      // display leaf name

      if (!::CertNameToStr(
         pClientCert->dwCertEncodingType,
         &pClientCert->pCertInfo->Subject,
         CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
         szName, sizeof(szName)))
      {
         const DWORD lastError = GetLastError();

         Output(_T("DisplayCertChain - Error building subject name: ") + GetLastErrorMessage(lastError));
      }

      Output(_T("Client subject: ") + _tstring(szName));


      if (!::CertNameToStr(
         pClientCert->dwCertEncodingType,
         &pClientCert->pCertInfo->Issuer,
         CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
         szName, sizeof(szName)))
      {
         const DWORD lastError = GetLastError();

         Output(_T("DisplayCertChain - Error building issuer name: ") + GetLastErrorMessage(lastError));
      }

      Output(_T("Client issuer: ") + _tstring(szName));

      // display certificate chain

      const CERT_CONTEXT *pCurrentCert = pClientCert;

      while (pCurrentCert)
      {
         DWORD dwVerificationFlags = 0;

         const CERT_CONTEXT *pIssuerCert = ::CertGetIssuerCertificateFromStore(
            pClientCert->hCertStore,
            pCurrentCert,
            nullptr,
            &dwVerificationFlags);

         if (!pIssuerCert)
         {
            if (pCurrentCert != pClientCert)
            {
               CertFreeCertificateContext(pCurrentCert);
            }
            break;
         }

         if (!::CertNameToStr(
            pIssuerCert->dwCertEncodingType,
            &pIssuerCert->pCertInfo->Subject,
            CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
            szName, sizeof(szName)))
         {
            const DWORD lastError = GetLastError();

            Output(_T("DisplayCertChain - Error building subject name: ") + GetLastErrorMessage(lastError));
         }

         Output(_T("CA subject: ") + _tstring(szName));

         if (!CertNameToStr(pIssuerCert->dwCertEncodingType,
            &pIssuerCert->pCertInfo->Issuer,
            CERT_X500_NAME_STR | CERT_NAME_STR_NO_PLUS_FLAG,
            szName, sizeof(szName)))
        {
            const DWORD lastError = GetLastError();

            Output(_T("DisplayCertChain - Error building issuer name: ") + GetLastErrorMessage(lastError));
         }

         Output(_T("CA issuer: ") + _tstring(szName));

         if (pCurrentCert != pClientCert)
         {
            CertFreeCertificateContext(pCurrentCert);
         }

         pCurrentCert = pIssuerCert;
         pIssuerCert = nullptr;
      }
   }
}

static HRESULT VerifyClientCertificate(
   const CERT_CONTEXT *pClientCert)
{
   if (!pClientCert)
   {
      return SEC_E_WRONG_PRINCIPAL;
   }

   //
   // Build certificate chain.
   //

   CERT_CHAIN_PARA  chainPara;

   LPSTR rgszUsages[] =
   {
      const_cast<LPSTR>(szOID_PKIX_KP_SERVER_AUTH),
      const_cast<LPSTR>(szOID_PKIX_KP_CLIENT_AUTH),
      const_cast<LPSTR>(szOID_SERVER_GATED_CRYPTO),
      const_cast<LPSTR>(szOID_SGC_NETSCAPE)
   };

   const DWORD cUsages = sizeof(rgszUsages) / sizeof(LPSTR);

   ZeroMemory(&chainPara, sizeof(chainPara));
   chainPara.cbSize = sizeof(chainPara);
   chainPara.RequestedUsage.dwType = USAGE_MATCH_TYPE_OR;
   chainPara.RequestedUsage.Usage.cUsageIdentifier = cUsages;
   chainPara.RequestedUsage.Usage.rgpszUsageIdentifier = rgszUsages;

   const CERT_CHAIN_CONTEXT *pChainContext = nullptr;

   if (!::CertGetCertificateChain(
      nullptr,
      pClientCert,
      nullptr,
      pClientCert->hCertStore,
      &chainPara,
      CERT_CHAIN_REVOCATION_CHECK_CHAIN,
      nullptr,
      &pChainContext))
   {
      return GetLastError();
   }

   //
   // Validate certificate chain.
   //

   HTTPSPolicyCallbackData  polHttps;

   const DWORD dwCertFlags = 0;

   ZeroMemory(&polHttps, sizeof(HTTPSPolicyCallbackData));
   polHttps.cbStruct = sizeof(HTTPSPolicyCallbackData);
   polHttps.dwAuthType = AUTHTYPE_CLIENT;
   polHttps.fdwChecks = dwCertFlags;
   polHttps.pwszServerName = nullptr;

   CERT_CHAIN_POLICY_PARA policyPara;

   memset(&policyPara, 0, sizeof(policyPara));
   policyPara.cbSize = sizeof(policyPara);
   policyPara.pvExtraPolicyPara = &polHttps;

   CERT_CHAIN_POLICY_STATUS policyStatus;

   memset(&policyStatus, 0, sizeof(policyStatus));
   policyStatus.cbSize = sizeof(policyStatus);

   DWORD status = SEC_E_OK;

   if (::CertVerifyCertificateChainPolicy(
      CERT_CHAIN_POLICY_SSL,
      pChainContext,
      &policyPara,
      &policyStatus))
   {
      if (policyStatus.dwError)
      {
         status = policyStatus.dwError;
      }
   }
   else
   {
      status = GetLastError();
   }

   if (pChainContext)
   {
      ::CertFreeCertificateChain(pChainContext);
   }

   return status;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
