///////////////////////////////////////////////////////////////////////////////
// File: ThreadPool.cpp
/////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2004 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "ThreadPool.h"
#include "SocketServer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"

#include "JetByteTools/IOTools/IBuffer.h"
#include "JetByteTools/IOTools/CustomPoolSmartBuffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::IThreadPoolWorkerThreadCallback;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::CCustomPoolSmartBuffer;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::CSmartStreamSocket;

///////////////////////////////////////////////////////////////////////////////
// CThreadPool
///////////////////////////////////////////////////////////////////////////////

CThreadPool::CThreadPool(
   const ThreadCount initialThreads,
   const ThreadCount minThreads,
   const ThreadCount maxThreads,
   const ThreadCount maxDormantThreads,
   const Milliseconds poolMaintPeriod,
   const Milliseconds dispatchTimeout)
   :  m_pServer(nullptr),
      m_threadPool(
         initialThreads,
         minThreads,
         maxThreads,
         maxDormantThreads,
         poolMaintPeriod,
         dispatchTimeout,
         *this)
{
}

void CThreadPool::Start(
   CSocketServer &server)
{
   if (m_pServer)
   {
      throw CException(_T("CThreadPool::Start()"), _T("Already started"));
   }

   m_pServer = &server;

   m_threadPool.Start();
}

void CThreadPool::WaitForShutdownToComplete(
   DWORD /*forceTerminationAfterMillis*/)
{
   m_threadPool.WaitForShutdownToComplete(/*forceTerminationAfterMillis*/);
}

void CThreadPool::DispatchCompleteMessage(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   if (!m_pServer)
   {
      throw CException(_T("CThreadPool::DispatchCompleteMessage()"), _T("Not started"));
   }

   socket.AddRef();

   m_threadPool.Dispatch(
      reinterpret_cast<ULONG_PTR>(&socket),
      0,
      buffer.Detach());
}

IThreadPoolWorkerThreadCallback &CThreadPool::CreateWorkerThread()
{
   return *this;
}

void CThreadPool::DestroyWorkerThread(
   IThreadPoolWorkerThreadCallback & /*thread*/)
{

}

void CThreadPool::DisposeOfQueuedItemsAfterShutdown(
   const ULONG_PTR completionKey,
   const DWORD /*operation*/,
   OVERLAPPED *pOverlapped)
{
   Output(_T("DisposeOfQueuedItemsAfterShutdown"));

   CSmartStreamSocket socket(reinterpret_cast<IStreamSocket *>(completionKey));

   CCustomPoolSmartBuffer buffer(static_cast<IBuffer *>(pOverlapped), socket.GetRef());

   m_pServer->ProcessMessageOnWorkerThread(socket.GetRef(), buffer, true);
}

bool CThreadPool::Initialise()
{
   return true;
}

void CThreadPool::Process(
   ULONG_PTR completionKey,
   DWORD operation,
   OVERLAPPED *pOverlapped)
{
   if (operation != 0)
   {
      throw CException(_T("CThreadPool::Process()"), _T("Unexpected operation: ") + ToString(operation));
   }

   CSmartStreamSocket socket(reinterpret_cast<IStreamSocket *>(completionKey));

   CCustomPoolSmartBuffer buffer(static_cast<IBuffer *>(pOverlapped), socket.GetRef());

   m_pServer->ProcessMessageOnWorkerThread(socket.GetRef(), buffer, false);
}

void CThreadPool::Shutdown()
{

}

///////////////////////////////////////////////////////////////////////////////
// End of file: ThreadPool.cpp
///////////////////////////////////////////////////////////////////////////////
