///////////////////////////////////////////////////////////////////////////////
// File: BufferAllocator.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2004 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "BufferAllocator.h"

#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/ToString.h"
#include "JetByteTools/CoreTools/Thread.h"

#pragma hdrstop

#include <set>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CThread;

#if JETBYTE_SERVER_COMMON_TRACK_BUFFER_ALLOCATION_USAGE_STATS == 1
using JetByteTools::Win32::CPerThreadUsageTracker;
#endif

using JetByteTools::IO::IMonitorBufferAllocation;

///////////////////////////////////////////////////////////////////////////////
// CBufferAllocator
///////////////////////////////////////////////////////////////////////////////

CBufferAllocator::CBufferAllocator(
   const BufferSize bufferSize,
   const BufferCount maxFreeBuffers,
   const bool log)
   :  JetByteTools::IO::CBufferAllocator(*this, bufferSize, maxFreeBuffers),
      m_log(log)
{

}

CBufferAllocator::CBufferAllocator(
   const BufferSize bufferSize,
   const BufferCount maxFreeBuffers,
   const AllocationFlags flags,
   const bool log)
   :  JetByteTools::IO::CBufferAllocator(*this, bufferSize, maxFreeBuffers, flags),
      m_log(log)
{

}

CBufferAllocator::~CBufferAllocator()
{
   // Must ensure that we have been flushed before we call the base class's
   // destructor otherwise it may try and call into us (as a monitor) when
   // it deletes buffers...

   Flush();

#if JETBYTE_SERVER_COMMON_TRACK_BUFFER_ALLOCATION_USAGE_STATS == 1

   OutputEx(_T("Buffer allocator usage stats"));

   CPerThreadUsageTracker::PerThreadStats allocations = m_perThreadAllocationStats.GetStats();

   CPerThreadUsageTracker::PerThreadStats releases = m_perThreadReleaseStats.GetStats();

   CPerThreadUsageTracker::PerThreadStats activity = m_perThreadActiveStats.GetStats();

   typedef std::set<DWORD> Threads;

   Threads threads;

   for (CPerThreadUsageTracker::PerThreadStats::const_iterator it = allocations.begin(), end = allocations.end();
      it != end;
      ++it)
   {
      threads.insert(it->first);
   }

   for (CPerThreadUsageTracker::PerThreadStats::const_iterator it = releases.begin(), end = releases.end();
      it != end;
      ++it)
   {
      threads.insert(it->first);
   }

   for (CPerThreadUsageTracker::PerThreadStats::const_iterator it = activity.begin(), end = activity.end();
      it != end;
      ++it)
   {
      threads.insert(it->first);
   }

   _tstring header = _T("Id - alloc - release - diff - max - min");

#if (JETBYTE_TRACK_THREAD_NAMES == 1)

   CThread::ThreadNames threadNames = CThread::GetThreadNames();

   header += _T(" - thread name");

#endif

   OutputEx(header);

   for (Threads::const_iterator it = threads.begin(), end = threads.end(); it != end; ++it)
   {
      const DWORD threadId = *it;

#if (JETBYTE_TRACK_THREAD_NAMES == 1)

      const _tstring threadName = _T(" - ") + threadNames[threadId];

#else

      const _tstring threadName;

#endif

      OutputEx(ToString(threadId) + _T(" - ") + ToString(allocations[threadId].total) + 
         _T(" - ") + ToString(releases[threadId].total) + 
         _T(" - ") + ToString(allocations[threadId].total - releases[threadId].total) + 
         _T(" - ") + ToString(activity[threadId].max) + 
         _T(" - ") + ToString(activity[threadId].min) + 
         threadName);
   }

#endif
}

void CBufferAllocator::OnBufferCreated()
{
   if (m_log)
   {
      Output(_T("OnBufferCreated"));
   }
}

void CBufferAllocator::OnBufferAllocated(
   IIndexedOpaqueUserData & /*userData*/)
{
#if JETBYTE_SERVER_COMMON_TRACK_BUFFER_ALLOCATION_USAGE_STATS == 1

   m_perThreadAllocationStats.Increment();
   m_perThreadActiveStats.Increment();

#endif

   if (m_log)
   {
      Output(_T("OnBufferAllocated"));
   }
}

void CBufferAllocator::OnBufferAllocationContention()
{
   if (m_log)
   {
      Output(_T("OnBufferAllocationContention"));
   }
}

void CBufferAllocator::OnBufferReleaseContention()
{
   if (m_log)
   {
      Output(_T("OnBufferReleaseContention"));
   }
}

void CBufferAllocator::OnBufferReleased(
   IIndexedOpaqueUserData & /*userData*/)
{
#if JETBYTE_SERVER_COMMON_TRACK_BUFFER_ALLOCATION_USAGE_STATS == 1

   m_perThreadActiveStats.Decrement();
   m_perThreadReleaseStats.Increment();

#endif

   if (m_log)
   {
      Output(_T("OnBufferReleased"));
   }
}

void CBufferAllocator::OnBufferDestroyed()
{
   if (m_log)
   {
      Output(_T("OnBufferDestroyed"));
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: BufferAllocator.cpp
///////////////////////////////////////////////////////////////////////////////
